# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import concurrent.futures
import contextlib
import ctypes
import datetime
import os
import queue
import socket
import struct
import sys
import threading
import time
import traceback
import urllib.parse
from typing import Callable, Optional, Union
from unittest import mock

import awscrt
import pytest
from awscrt.http import HttpProxyOptions, HttpRequest
from awscrt.io import ClientBootstrap, SocketOptions, TlsConnectionOptions
from awscrt.websocket import (
    IncomingFrame,
    OnConnectionSetupData,
    OnConnectionShutdownData,
    OnIncomingFrameBeginData,
    OnIncomingFrameCompleteData,
    OnIncomingFramePayloadData,
    OnSendFrameCompleteData,
    Opcode,
)
from dateutil.tz import tzutc

from awscli.compat import is_windows
from awscli.customizations.ec2instanceconnect.websocket import (
    BaseWebsocketIO,
    WebsocketManager,
)
from tests import AWSRequest, CLIRunner, HTTPResponse, SessionStubber

STDIN_QUEUE = queue.Queue()
STDOUT_QUEUE = queue.Queue()
SHUTDOWN_SENTINEL = object()


class StubStdinStdoutIO(BaseWebsocketIO):
    def has_data_to_read(self):
        return not STDIN_QUEUE.empty()

    def read(self, amt) -> bytes:
        return STDIN_QUEUE.get()

    def write(self, data):
        STDOUT_QUEUE.put(data)

    def close(self):
        pass


def pop_stdout_content():
    return STDOUT_QUEUE.get(block=False)


def assert_stdout_empty():
    assert STDOUT_QUEUE.empty()


class MockCRTWebsocket(threading.Thread):
    def __init__(self):
        super().__init__()
        self.url = None
        self.client_input_queue = queue.Queue()
        self._server_input_queue = queue.Queue()
        self._shutdown_event = threading.Event()

        # Callbacks that get set when connect is called
        self._on_connection_setup = None
        self._on_connection_shutdown = None
        self._on_incoming_frame_begin = None
        self._on_incoming_frame_payload = None
        self._on_incoming_frame_complete = None

        self._shutdown_reason = None
        self._shutdown_exception = None

    def connect(
        self,
        *,
        host: str,
        port: Optional[int] = None,
        handshake_request: HttpRequest,
        bootstrap: Optional[ClientBootstrap] = None,
        socket_options: Optional[SocketOptions] = None,
        tls_connection_options: Optional[TlsConnectionOptions] = None,
        proxy_options: Optional[HttpProxyOptions] = None,
        manage_read_window: bool = False,
        initial_read_window: Optional[int] = None,
        on_connection_setup: Callable[[OnConnectionSetupData], None],
        on_connection_shutdown: Optional[
            Callable[[OnConnectionShutdownData], None]
        ] = None,
        on_incoming_frame_begin: Optional[
            Callable[[OnIncomingFrameBeginData], None]
        ] = None,
        on_incoming_frame_payload: Optional[
            Callable[[OnIncomingFramePayloadData], None]
        ] = None,
        on_incoming_frame_complete: Optional[
            Callable[[OnIncomingFrameCompleteData], None]
        ] = None,
    ):
        self.url = host + handshake_request.path

        self._on_connection_setup = on_connection_setup
        self._on_connection_shutdown = on_connection_shutdown
        self._on_incoming_frame_begin = on_incoming_frame_begin
        self._on_incoming_frame_payload = on_incoming_frame_payload
        self._on_incoming_frame_complete = on_incoming_frame_complete

        crt_websocket = mock.Mock(awscrt.websocket.WebSocket)
        crt_websocket.send_frame = self.send_frame_from_client
        data = OnConnectionSetupData()
        data.handshake_response_headers = []
        data.websocket = crt_websocket
        self._on_connection_setup(data)

        # Kick off thread that mocks reading and writing to the client.
        self.start()

    # Appends data to the server_input_queue, which will represent the server sending data to the websocket
    def add_output_from_server(self, data_in_bytes):
        self._server_input_queue.put(data_in_bytes)

    # Appends data to the client_input_queue, which will represent the client sending data to the websocket
    def send_frame_from_client(
        self,
        opcode: Opcode,
        payload: Optional[Union[str, bytes, bytearray, memoryview]] = None,
        *,
        fin: bool = True,
        on_complete: Optional[
            Callable[[OnSendFrameCompleteData], None]
        ] = None,
    ):
        if opcode == Opcode.BINARY:
            data = OnSendFrameCompleteData()
            on_complete(data)
            if len(payload) > 0:
                self.client_input_queue.put(payload)

    def add_shutdown_from_server(
        self, shutdown_reason=None, shutdown_exception=None
    ):
        self._server_input_queue.put(SHUTDOWN_SENTINEL)
        self._shutdown_reason = shutdown_reason
        self._shutdown_exception = shutdown_exception

    def run(self):
        while not self._shutdown_event.is_set():
            if not self._server_input_queue.empty():
                data_in_bytes = self._server_input_queue.get()
                if data_in_bytes == SHUTDOWN_SENTINEL:
                    self._shutdown_event.set()
                else:
                    incoming_frame = IncomingFrame(
                        opcode=Opcode.BINARY,
                        payload_length=len(data_in_bytes),
                        fin=True,
                    )
                    payload_data = OnIncomingFramePayloadData(
                        frame=incoming_frame, data=data_in_bytes
                    )
                    self._on_incoming_frame_payload(payload_data)

                    complete_data = OnIncomingFrameCompleteData(
                        frame=incoming_frame
                    )
                    self._on_incoming_frame_complete(complete_data)

        if self._shutdown_reason:
            shutdown_code = 1
            packed_shutdown_code = struct.pack(">H", shutdown_code)
            encoded_shutdown_reason = self._shutdown_reason.encode()
            data_in_bytes = packed_shutdown_code + encoded_shutdown_reason
            incoming_frame = IncomingFrame(
                opcode=Opcode.CLOSE,
                payload_length=len(data_in_bytes),
                fin=True,
            )
            payload_data = OnIncomingFramePayloadData(
                frame=incoming_frame, data=data_in_bytes
            )
            self._on_incoming_frame_payload(payload_data)

            complete_data = OnIncomingFrameCompleteData(frame=incoming_frame)
            self._on_incoming_frame_complete(complete_data)

        self._on_connection_shutdown(
            OnConnectionShutdownData(exception=self._shutdown_exception)
        )


@pytest.fixture
def session_stubber():
    return SessionStubber()


@pytest.fixture
def region():
    return "us-west-2"


@pytest.fixture
def cli_runner(session_stubber, region):
    cli_runner = CLIRunner(session_stubber=session_stubber)
    cli_runner.env["AWS_DEFAULT_REGION"] = region
    return cli_runner


@pytest.fixture(autouse=True)
def clean_io_queues():
    while not STDIN_QUEUE.empty():
        STDIN_QUEUE.get()
    while not STDOUT_QUEUE.empty():
        STDOUT_QUEUE.get()


@pytest.fixture
def io_patch():
    io_to_patch = "WindowsStdinStdoutIO" if is_windows else "StdinStdoutIO"
    new_websocket_io = StubStdinStdoutIO
    return mock.patch(
        f"awscli.customizations.ec2instanceconnect.websocket.{io_to_patch}",
        new=new_websocket_io,
    )


@pytest.fixture
def mock_crt_websocket():
    return MockCRTWebsocket()


@pytest.fixture
def connect_patch(mock_crt_websocket):
    return mock.patch(
        "awscrt.websocket.connect", new=mock_crt_websocket.connect
    )


@pytest.fixture
def describe_instance_response():
    return """
    <DescribeInstancesResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
        <reservationSet>
            <item>
                <instancesSet>
                    <item>
                        <instanceId>i-123</instanceId>
                        <subnetId>subnet-123</subnetId>
                        <vpcId>vpc-123</vpcId>
                        <privateIpAddress>10.0.0.0</privateIpAddress>
                    </item>
                </instancesSet>
            </item>
        </reservationSet>
    </DescribeInstancesResponse>
    """


@pytest.fixture
def request_params_for_describe_instance():
    return {'InstanceIds': ['i-123']}


@pytest.fixture
def dns_name():
    return "eice-123.ec2-instance-connect-endpoint.us-west-2.amazonaws.com"


@pytest.fixture
def fips_dns_name():
    return (
        "eice-123.fips.ec2-instance-connect-endpoint.us-west-2.amazonaws.com"
    )


@pytest.fixture
def describe_eice_response(dns_name, fips_dns_name):
    return f"""
    <DescribeInstanceConnectEndpointsResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
        <instanceConnectEndpointSet>
            <item>
                <dnsName>{dns_name}</dnsName>
                <fipsDnsName>{fips_dns_name}</fipsDnsName>
                <instanceConnectEndpointId>eice-123</instanceConnectEndpointId>
                <state>create-complete</state>
                <subnetId>subnet-123</subnetId>
                <vpcId>vpc-123</vpcId>
            </item>
        </instanceConnectEndpointSet>
    </DescribeInstanceConnectEndpointsResponse>
    """


@pytest.fixture
def describe_eice_response_without_fips_dns(dns_name, fips_dns_name):
    return f"""
    <DescribeInstanceConnectEndpointsResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
        <instanceConnectEndpointSet>
            <item>
                <dnsName>{dns_name}</dnsName>
                <instanceConnectEndpointId>eice-123</instanceConnectEndpointId>
                <state>create-complete</state>
                <subnetId>subnet-123</subnetId>
                <vpcId>vpc-123</vpcId>
            </item>
        </instanceConnectEndpointSet>
    </DescribeInstanceConnectEndpointsResponse>
    """


@pytest.fixture
def describe_eice_response_empty_fips_dns(dns_name, fips_dns_name):
    return f"""
    <DescribeInstanceConnectEndpointsResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
        <instanceConnectEndpointSet>
            <item>
                <dnsName>{dns_name}</dnsName>
                <fipsDnsName></fipsDnsName>
                <instanceConnectEndpointId>eice-123</instanceConnectEndpointId>
                <state>create-complete</state>
                <subnetId>subnet-123</subnetId>
                <vpcId>vpc-123</vpcId>
            </item>
        </instanceConnectEndpointSet>
    </DescribeInstanceConnectEndpointsResponse>
    """


@pytest.fixture
def request_params_for_describe_eice():
    return {
        'Filters': [
            {
                'Name': 'state',
                'Values': [
                    'create-complete',
                    'update-in-progress',
                    'update-failed',
                    'update-complete',
                ],
            },
            {'Name': 'vpc-id', 'Values': ['vpc-123']},
        ]
    }


@pytest.fixture
def datetime_utcnow_patch():
    clock = datetime.datetime(2020, 1, 1, 1, 1, 1, tzinfo=tzutc())
    with mock.patch('datetime.datetime') as dt:
        dt.utcnow.return_value = clock
        yield dt


@pytest.fixture
def run_listener_cmd_in_thread_executor(
    cli_runner, mock_crt_websocket, connect_patch, io_patch
):
    stop_event = threading.Event()
    with mock.patch.object(
        WebsocketManager, 'RUNNING', new_callable=mock.PropertyMock
    ) as websocket_join:
        websocket_join.return_value = stop_event
        with connect_patch, io_patch:
            with concurrent.futures.ThreadPoolExecutor() as executor:

                def submit_run_command(cmdline):
                    return executor.submit(cli_runner.run, cmdline)

                yield submit_run_command
                stop_event.set()


def connect_to_listener(listener_future):
    retry_connection = True
    attempts = 0
    max_attempts = 5
    tracebacks = ''
    while retry_connection:
        attempts += 1
        time.sleep(0.2)
        try:
            return socket.create_connection(("localhost", 3333), timeout=5)
        except (ConnectionError, OSError):
            tracebacks += f'Traceback from attempt {attempts}:\n{traceback.format_exc()}\n'
            if attempts == max_attempts:
                retry_connection = False
        except Exception:
            tracebacks += f'Traceback from attempt {attempts}:\n{traceback.format_exc()}\n'
            retry_connection = False
    msg = (
        f'Failed to connect to open-tunnel command after {attempts} attempts.\n'
        f'{tracebacks}'
    )
    if listener_future.done():
        cli_runner_result = listener_future.result()
        msg += (
            f'RC from open-tunnel cmd: {cli_runner_result.rc}\n'
            f'Stdout from open-tunnel cmd: {cli_runner_result.stdout}\n'
            f'Stderr from open-tunnel cmd: {cli_runner_result.stderr}'
        )
    assert False, msg


def assert_url(dns_name, url):
    _, _, path, _, qs, _ = urllib.parse.urlparse(url)
    parsed_qs = urllib.parse.parse_qs(qs)

    assert path == f"{dns_name}/openTunnel"
    assert parsed_qs["instanceConnectEndpointId"]
    assert parsed_qs["remotePort"]
    assert parsed_qs["privateIpAddress"]
    assert parsed_qs["maxTunnelDuration"]
    assert parsed_qs["X-Amz-Algorithm"]
    assert parsed_qs["X-Amz-Credential"]
    assert parsed_qs["X-Amz-Date"]
    assert parsed_qs["X-Amz-SignedHeaders"]
    assert parsed_qs["X-Amz-Signature"]
    assert parsed_qs["X-Amz-Expires"] == ["60"]


class TestOpenTunnel:
    @pytest.mark.parametrize(
        "endpoint_state",
        [
            "create-complete",
            "update-in-progress",
            "update-failed",
            "update-complete",
        ],
    )
    def test_single_connection_mode(
        self,
        cli_runner,
        mock_crt_websocket,
        connect_patch,
        io_patch,
        describe_instance_response,
        dns_name,
        fips_dns_name,
        request_params_for_describe_instance,
        request_params_for_describe_eice,
        datetime_utcnow_patch,
        endpoint_state,
    ):
        cli_runner.env["AWS_USE_FIPS_ENDPOINT"] = "false"
        cmdline = [
            "ec2-instance-connect",
            "open-tunnel",
            "--instance-id",
            "i-123",
            "--max-tunnel-duration",
            "1",
        ]

        # Create endpoint response with the specified state
        describe_eice_response_with_state = f"""
        <DescribeInstanceConnectEndpointsResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
            <instanceConnectEndpointSet>
                <item>
                    <dnsName>{dns_name}</dnsName>
                    <fipsDnsName>{fips_dns_name}</fipsDnsName>
                    <instanceConnectEndpointId>eice-123</instanceConnectEndpointId>
                    <state>{endpoint_state}</state>
                    <subnetId>subnet-123</subnetId>
                    <vpcId>vpc-123</vpcId>
                </item>
            </instanceConnectEndpointSet>
        </DescribeInstanceConnectEndpointsResponse>
        """

        cli_runner.add_response(HTTPResponse(body=describe_instance_response))
        cli_runner.add_response(
            HTTPResponse(body=describe_eice_response_with_state)
        )

        test_server_input = b"Test Server Output"
        mock_crt_websocket.add_output_from_server(test_server_input)
        mock_crt_websocket.add_shutdown_from_server()

        with connect_patch, io_patch:
            result = cli_runner.run(cmdline)

        assert 0 == result.rc
        assert pop_stdout_content() == test_server_input
        assert_stdout_empty()
        # Order of the query params on the url matter because of sigv4
        assert (
            "eice-123.ec2-instance-connect-endpoint.us-west-2.amazonaws.com/openTunnel?"
            "instanceConnectEndpointId=eice-123&remotePort=22&privateIpAddress=10.0.0.0&"
            "maxTunnelDuration=1&X-Amz-Algorithm=AWS4-HMAC-SHA256&"
            "X-Amz-Credential=access_key%2F20200101%2Fus-west-2%2Fec2-instance-connect%2Faws4_request&"
            "X-Amz-Date=20200101T010101Z&X-Amz-Expires=60&X-Amz-SignedHeaders=host&"
            "X-Amz-Signature=3a56740422a5b9aebd22d860b7a30c729e459a6f71a83bc0de3a2b44b7353f28"
            == mock_crt_websocket.url
        )
        assert_url(dns_name, mock_crt_websocket.url)
        assert result.aws_requests == [
            AWSRequest(
                service_name='ec2',
                operation_name='DescribeInstances',
                params=request_params_for_describe_instance,
            ),
            AWSRequest(
                service_name='ec2',
                operation_name='DescribeInstanceConnectEndpoints',
                params=request_params_for_describe_eice,
            ),
        ]

    def test_single_connection_mode_with_no_describe_calls(
        self,
        cli_runner,
        mock_crt_websocket,
        connect_patch,
        io_patch,
        dns_name,
    ):
        cmdline = [
            "ec2-instance-connect",
            "open-tunnel",
            "--private-ip-address",
            "10.0.0.0",
            "--instance-connect-endpoint-id",
            "eice-123",
            "--instance-connect-endpoint-dns-name",
            dns_name,
            "--max-tunnel-duration",
            "1",
        ]

        test_server_input = b"Test Server Output"
        mock_crt_websocket.add_output_from_server(test_server_input)
        mock_crt_websocket.add_shutdown_from_server()

        with connect_patch, io_patch:
            result = cli_runner.run(cmdline)

        assert 0 == result.rc
        assert pop_stdout_content() == test_server_input
        assert_stdout_empty()
        assert_url(dns_name, mock_crt_websocket.url)
        assert result.aws_requests == []

    def test_multiple_connection_mode(
        self,
        run_listener_cmd_in_thread_executor,
        mock_crt_websocket,
    ):
        cmdline = [
            "ec2-instance-connect",
            "open-tunnel",
            "--instance-id",
            "i-123",
            "--private-ip-address",
            "10.0.0.0",
            "--instance-connect-endpoint-id",
            "eice-123",
            "--instance-connect-endpoint-dns-name",
            "test",
            "--local-port",
            "3333",
        ]
        future = run_listener_cmd_in_thread_executor(cmdline)
        # Retry several times since it can take few ms to set up everything
        with contextlib.closing(connect_to_listener(future)) as s:
            test_server_input = b"Test Server Output"
            mock_crt_websocket.add_output_from_server(test_server_input)

            # Reading from the TCP Conn should return test server input
            data = s.recv(1024)
            assert test_server_input == data

            test_client_input = b"Test Client Input"
            # Writing to the TCP Conn should ensure a frame is sent with that data
            s.sendall(test_client_input)
            client_input_sent_to_websocket = (
                mock_crt_websocket.client_input_queue.get(timeout=5)
            )
            assert test_client_input == client_input_sent_to_websocket

            mock_crt_websocket.add_shutdown_from_server()

    def test_command_uses_fips_endpoint_to_connect(
        self,
        cli_runner,
        mock_crt_websocket,
        connect_patch,
        io_patch,
        describe_instance_response,
        fips_dns_name,
        describe_eice_response,
    ):
        cli_runner.env["AWS_USE_FIPS_ENDPOINT"] = "true"
        cmdline = [
            "ec2-instance-connect",
            "open-tunnel",
            "--instance-id",
            "i-123",
            "--max-tunnel-duration",
            "1",
        ]
        cli_runner.add_response(HTTPResponse(body=describe_instance_response))
        cli_runner.add_response(HTTPResponse(body=describe_eice_response))

        test_server_input = b"Test Server Output"
        mock_crt_websocket.add_output_from_server(test_server_input)
        mock_crt_websocket.add_shutdown_from_server()

        with connect_patch, io_patch:
            result = cli_runner.run(cmdline)

        assert 0 == result.rc
        assert_url(fips_dns_name, mock_crt_websocket.url)

    def test_command_returns_shutdown_reason(
        self, cli_runner, mock_crt_websocket, connect_patch, io_patch
    ):
        cmdline = [
            "ec2-instance-connect",
            "open-tunnel",
            "--instance-id",
            "i-123",
            "--private-ip-address",
            "10.0.0.0",
            "--instance-connect-endpoint-id",
            "eice-123",
            "--instance-connect-endpoint-dns-name",
            "test",
        ]

        shutdown_reason = "Test Shutdown"
        mock_crt_websocket.add_shutdown_from_server(
            shutdown_reason=shutdown_reason
        )

        with connect_patch, io_patch:
            result = cli_runner.run(cmdline)

        assert 255 == result.rc
        assert shutdown_reason in result.stderr

    def test_command_returns_shutdown_exception(
        self, cli_runner, mock_crt_websocket, connect_patch, io_patch
    ):
        cmdline = [
            "ec2-instance-connect",
            "open-tunnel",
            "--instance-id",
            "i-123",
            "--private-ip-address",
            "10.0.0.0",
            "--instance-connect-endpoint-id",
            "eice-123",
            "--instance-connect-endpoint-dns-name",
            "test",
        ]

        shutdown_exception_message = "Test Shutdown Error"
        mock_crt_websocket.add_shutdown_from_server(
            shutdown_exception=RuntimeError(shutdown_exception_message)
        )

        with connect_patch, io_patch:
            result = cli_runner.run(cmdline)

        assert 255 == result.rc
        assert shutdown_exception_message in result.stderr

    @pytest.mark.parametrize(
        "cli_input,expected",
        [
            # Customer must provide instance id or private ip
            (
                [
                    "ec2-instance-connect",
                    "open-tunnel",
                    "--instance-connect-endpoint-id",
                    "eice-123",
                    "--instance-connect-endpoint-dns-name",
                    "test",
                ],
                "Specify an instance id or private ip",
            ),
            # Customer must define eice id when dns name is provided
            (
                [
                    "ec2-instance-connect",
                    "open-tunnel",
                    "--instance-id",
                    "i-123",
                    "--instance-connect-endpoint-dns-name",
                    "test",
                ],
                "Specify an instance connect endpoint id",
            ),
            # Customer must define eice id when private ip is provided
            (
                [
                    "ec2-instance-connect",
                    "open-tunnel",
                    "--instance-id",
                    "i-123",
                    "--private-ip-address",
                    "0.0.0.0",
                ],
                "Specify an instance connect endpoint id",
            ),
            # Customer must use provide local port when using command in isatty mode
            (
                [
                    "ec2-instance-connect",
                    "open-tunnel",
                    "--instance-id",
                    "i-123",
                ],
                "This command does not support interactive mode",
            ),
        ],
    )
    @mock.patch("sys.stdin")
    def test_command_fails_when_invalid_input(
        self, mock_stdin, cli_runner, cli_input, expected
    ):
        mock_stdin.isatty.return_value = True
        result = cli_runner.run(cli_input)

        assert 252 == result.rc
        assert expected in result.stderr

    @pytest.mark.parametrize("duration", ["-1", "0", "3601"])
    def test_command_fails_when_using_invalid_max_tunnel_duration(
        self, cli_runner, duration
    ):
        cmdline = [
            "ec2-instance-connect",
            "open-tunnel",
            "--instance-id",
            "i-123",
            "--max-tunnel-duration",
            duration,
        ]
        result = cli_runner.run(cmdline)

        assert 252 == result.rc
        assert "Invalid max connection timeout specified" in result.stderr

    def test_command_fails_when_no_fips_endpoint_available_to_connect(
        self,
        cli_runner,
        describe_instance_response,
        describe_eice_response_without_fips_dns,
    ):
        cli_runner.env["AWS_USE_FIPS_ENDPOINT"] = "true"
        cmdline = [
            "ec2-instance-connect",
            "open-tunnel",
            "--instance-id",
            "i-123",
            "--max-tunnel-duration",
            "1",
        ]
        cli_runner.add_response(HTTPResponse(body=describe_instance_response))
        cli_runner.add_response(
            HTTPResponse(body=describe_eice_response_without_fips_dns)
        )

        result = cli_runner.run(cmdline)

        assert 253 == result.rc
        assert "Unable to find FIPS Endpoint" in result.stderr

    def test_command_fails_when_empty_fips_endpoint_available_to_connect(
        self,
        cli_runner,
        describe_instance_response,
        describe_eice_response_empty_fips_dns,
    ):
        cli_runner.env["AWS_USE_FIPS_ENDPOINT"] = "true"
        cmdline = [
            "ec2-instance-connect",
            "open-tunnel",
            "--instance-id",
            "i-123",
            "--max-tunnel-duration",
            "1",
        ]
        cli_runner.add_response(HTTPResponse(body=describe_instance_response))
        cli_runner.add_response(
            HTTPResponse(body=describe_eice_response_empty_fips_dns)
        )

        result = cli_runner.run(cmdline)

        assert 253 == result.rc
        assert "Unable to find FIPS Endpoint" in result.stderr



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import copy
import json
import os

from awscli.customizations.emr.emrfsutils import (
    CONSISTENT_OPTION_NAME,
    CSE_CUSTOM_OPTION_NAME,
    CSE_KMS_OPTION_NAME,
    CSE_OPTION_NAME,
)
from tests.unit.customizations.emr import (
    EMRBaseAWSCommandParamsTest as BaseAWSCommandParamsTest,
)

DEFAULT_INSTANCES = {
    'KeepJobFlowAliveWhenNoSteps': True,
    'TerminationProtected': False,
    'InstanceGroups': [
        {
            'InstanceRole': 'MASTER',
            'InstanceCount': 1,
            'Name': 'MASTER',
            'Market': 'ON_DEMAND',
            'InstanceType': 'm1.large',
        }
    ],
}

DEFAULT_CMD = (
    'emr create-cluster --use-default-roles' ' --instance-type m1.large '
)
DEFAULT_RESULT = {
    'Name': "Development Cluster",
    'Instances': DEFAULT_INSTANCES,
    'VisibleToAllUsers': True,
    'JobFlowRole': "EMR_EC2_DefaultRole",
    'ServiceRole': "EMR_DefaultRole",
    'Tags': [],
}

EMPTY_EMRFS_CONFIGURATION = {'Classification': 'emrfs-site', 'Properties': {}}

DEFAULT_CONFIGURATIONS = [
    {
        'Classification': 'hadoop-env',
        'Configurations': [],
        'Properties': {'someProperty': 'someValue'},
    }
]


class TestEmrfsUtils(BaseAWSCommandParamsTest):
    def test_consistent(self):
        emrfs_option_value = 'Consistent=true'
        expected_emrfs_properties = {'fs.s3.consistent': 'true'}
        expected_emrfs_ba_key_values = ['fs.s3.consistent=true']

        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_consistent_w_optional_args(self):
        emrfs_option_value = 'Consistent=true,RetryCount=5,RetryPeriod=30'

        expected_emrfs_properties = {
            'fs.s3.consistent': 'true',
            'fs.s3.consistent.retryCount': '5',
            'fs.s3.consistent.retryPeriodSeconds': '30',
        }

        expected_emrfs_ba_key_values = [
            'fs.s3.consistent=true',
            'fs.s3.consistent.retryCount=5',
            'fs.s3.consistent.retryPeriodSeconds=30',
        ]

        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_consistent_false_w_optional_args(self):
        emrfs_option_value = 'Consistent=false,RetryCount=5'

        expected_emrfs_properties = {
            'fs.s3.consistent': 'false',
            'fs.s3.consistent.retryCount': '5',
        }

        expected_emrfs_ba_key_values = [
            'fs.s3.consistent=false',
            'fs.s3.consistent.retryCount=5',
        ]

        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_sse(self):
        emrfs_option_value = 'SSE=true'
        expected_emrfs_ba_key_values = [
            'fs.s3.enableServerSideEncryption=true'
        ]
        expected_emrfs_properties = {
            'fs.s3.enableServerSideEncryption': 'true'
        }
        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

        emrfs_option_value = 'Encryption=ServerSide'
        expected_emrfs_ba_key_values = [
            'fs.s3.enableServerSideEncryption=true'
        ]
        expected_emrfs_properties = {
            'fs.s3.enableServerSideEncryption': 'true'
        }
        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_cse_kms(self):
        emrfs_option_value = (
            'Encryption=ClientSide,ProviderType=KMS,' 'KMSKeyId=my_key'
        )
        expected_emrfs_ba_key_values = [
            'fs.s3.cse.enabled=true',
            'fs.s3.cse.encryptionMaterialsProvider='
            'com.amazon.ws.emr.hadoop.fs.cse.KMSEncryptionMaterialsProvider',
            'fs.s3.cse.kms.keyId=my_key',
        ]
        expected_emrfs_properties = {
            'fs.s3.cse.enabled': 'true',
            'fs.s3.cse.encryptionMaterialsProvider': 'com.amazon.ws.emr.hadoop.fs.cse.'
            'KMSEncryptionMaterialsProvider',
            'fs.s3.cse.kms.keyId': 'my_key',
        }
        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_cse_custom(self):
        emrfs_option_value = (
            'Encryption=ClientSide,ProviderType=Custom,'
            'CustomProviderLocation=my_location,CustomProviderClass=my_class'
        )
        expected_emrfs_ba_key_values = [
            'fs.s3.cse.enabled=true',
            'fs.s3.cse.encryptionMaterialsProvider=' 'my_class',
        ]
        expected_emrfs_properties = {
            'fs.s3.cse.enabled': 'true',
            'fs.s3.cse.encryptionMaterialsProvider': 'my_class',
            'fs.s3.cse.encryptionMaterialsProvider.uri': 'my_location',
        }

        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
            'my_location',
        )

    def test_sse_and_consistent(self):
        emrfs_option_value = 'SSE=true,Consistent=true'
        expected_emrfs_ba_key_values = [
            'fs.s3.consistent=true',
            'fs.s3.enableServerSideEncryption=true',
        ]
        expected_emrfs_properties = {
            'fs.s3.consistent': 'true',
            'fs.s3.enableServerSideEncryption': 'true',
        }
        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

        emrfs_option_value = 'Consistent=false,Encryption=serVERSIde'
        expected_emrfs_ba_key_values = [
            'fs.s3.consistent=false',
            'fs.s3.enableServerSideEncryption=true',
        ]
        expected_emrfs_properties = {
            'fs.s3.consistent': 'false',
            'fs.s3.enableServerSideEncryption': 'true',
        }

        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_cse_and_consistent(self):
        emrfs_option_value = (
            'Encryption=ClientSide,ProviderType=KMS,'
            'KMSKeyId=my_key,Consistent=true'
        )
        expected_emrfs_ba_key_values = [
            'fs.s3.consistent=true',
            'fs.s3.cse.enabled=true',
            'fs.s3.cse.encryptionMaterialsProvider=com.amazon.ws.emr.'
            'hadoop.fs.cse.KMSEncryptionMaterialsProvider',
            'fs.s3.cse.kms.keyId=my_key',
        ]
        expected_emrfs_properties = {
            'fs.s3.consistent': 'true',
            'fs.s3.cse.enabled': 'true',
            'fs.s3.cse.encryptionMaterialsProvider': 'com.amazon.ws.emr.'
            'hadoop.fs.cse.KMSEncryptionMaterialsProvider',
            'fs.s3.cse.kms.keyId': 'my_key',
        }

        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_args_and_sse(self):
        emrfs_option_value = (
            'SSE=true,Args=[fs.s3.serverSideEncryptionAlgorithm=AES256]'
        )
        expected_emrfs_ba_key_values = [
            'fs.s3.enableServerSideEncryption=true',
            'fs.s3.serverSideEncryptionAlgorithm=AES256',
        ]
        expected_emrfs_properties = {
            'fs.s3.enableServerSideEncryption': 'true',
            'fs.s3.serverSideEncryptionAlgorithm': 'AES256',
        }

        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_args_and_cse(self):
        emrfs_option_value = (
            'Encryption=ClientSide,ProviderType=KMS,'
            'KMSKeyId=my_key,Args=[k1=v1]'
        )
        expected_emrfs_ba_key_values = [
            'fs.s3.cse.enabled=true',
            'fs.s3.cse.encryptionMaterialsProvider=com.amazon.ws.emr.'
            'hadoop.fs.cse.KMSEncryptionMaterialsProvider',
            'fs.s3.cse.kms.keyId=my_key',
            'k1=v1',
        ]
        expected_emrfs_properties = {
            'fs.s3.cse.enabled': 'true',
            'fs.s3.cse.encryptionMaterialsProvider': 'com.amazon.ws.emr.'
            'hadoop.fs.cse.KMSEncryptionMaterialsProvider',
            'fs.s3.cse.kms.keyId': 'my_key',
            'k1': 'v1',
        }

        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_args_and_consistent(self):
        emrfs_option_value = 'Consistent=true,Args=[k1=v1,k2=v2]'
        expected_emrfs_ba_key_values = [
            'fs.s3.consistent=true',
            'k1=v1',
            'k2=v2',
        ]
        expected_emrfs_properties = {
            'fs.s3.consistent': 'true',
            'k1': 'v1',
            'k2': 'v2',
        }

        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_only_args(self):
        emrfs_option_value = 'Args=[k1=v1,k2=v2,k3]'
        expected_emrfs_ba_key_values = ['k1=v1', 'k2=v2', 'k3']
        expected_emrfs_properties = {'k1': 'v1', 'k2': 'v2', 'k3': ''}

        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_using_json_file(self):
        data_path = os.path.join(
            os.path.dirname(__file__), 'input_emr_fs.json'
        )
        emrfs_option_value = 'file://%s' % data_path
        expected_emrfs_ba_key_values = [
            'fs.s3.consistent=true',
            'fs.s3.consistent.retryCount=10',
            'fs.s3.consistent.retryPeriodSeconds=3',
            'fs.s3.enableServerSideEncryption=false',
            'fs.s3.serverSideEncryptionAlgorithm=AES256',
            'fs.s3.sleepTimeSeconds=30',
        ]
        expected_emrfs_properties = {
            'fs.s3.consistent': 'true',
            'fs.s3.consistent.retryCount': '10',
            'fs.s3.consistent.retryPeriodSeconds': '3',
            'fs.s3.enableServerSideEncryption': 'false',
            'fs.s3.serverSideEncryptionAlgorithm': 'AES256',
            'fs.s3.sleepTimeSeconds': '30',
        }

        self._assert_bootstrap_actions(
            emrfs_option_value,
            expected_emrfs_ba_key_values,
            expected_emrfs_properties,
        )

    def test_only_one_encryption_type(self):
        self._assert_error_msg(
            emrfs_option_value='SSE=true,Encryption=ClientSide,'
            'ProviderType=KMS,KMSKeyId=k1',
            exception_class_name='BothSseAndEncryptionConfiguredError',
            error_msg_kwargs={'sse': 'True', 'encryption': 'ClientSide'},
            rc=252,
        )

    def test_cse_missing_provider_type(self):
        self._assert_error_msg(
            emrfs_option_value='Encryption=ClientSide',
            exception_class_name='MissingParametersError',
            error_msg_kwargs={
                'object_name': CSE_OPTION_NAME,
                'missing': 'ProviderType',
            },
            rc=252,
        )

    def test_cse_kms_missing_key_id(self):
        self._assert_error_msg(
            emrfs_option_value='Encryption=ClientSide,ProviderType=KMS',
            exception_class_name='MissingParametersError',
            error_msg_kwargs={
                'object_name': CSE_KMS_OPTION_NAME,
                'missing': 'KMSKeyId',
            },
            rc=252,
        )

    def test_cse_custom_missing_all(self):
        self._assert_error_msg(
            emrfs_option_value='Encryption=ClientSide,ProviderType=Custom',
            exception_class_name='MissingParametersError',
            error_msg_kwargs={
                'object_name': CSE_CUSTOM_OPTION_NAME,
                'missing': 'CustomProviderClass and ' 'CustomProviderLocation',
            },
            rc=252,
        )

    def test_cse_custom_missing_provider_class(self):
        self._assert_error_msg(
            emrfs_option_value='Encryption=ClientSide,ProviderType=Custom,'
            'CustomProviderLocation=my_location',
            exception_class_name='MissingParametersError',
            error_msg_kwargs={
                'object_name': CSE_CUSTOM_OPTION_NAME,
                'missing': 'CustomProviderClass',
            },
            rc=252,
        )

    def test_cse_custom_missing_provider_location(self):
        self._assert_error_msg(
            emrfs_option_value='Encryption=ClientSide,ProviderType=Custom,'
            'CustomProviderClass=my_class',
            exception_class_name='MissingParametersError',
            error_msg_kwargs={
                'object_name': CSE_CUSTOM_OPTION_NAME,
                'missing': 'CustomProviderLocation',
            },
            rc=252,
        )

    def test_valid_encryption(self):
        self._assert_error_msg(
            emrfs_option_value='Encryption=ClientSide1',
            exception_class_name='UnknownEncryptionTypeError',
            error_msg_kwargs={'encryption': 'ClientSide1'},
            rc=252,
        )

    def test_valid_cse_provider_type(self):
        self._assert_error_msg(
            emrfs_option_value='Encryption=ClientSide,ProviderType=KMS1',
            exception_class_name='UnknownCseProviderTypeError',
            error_msg_kwargs={'provider_type': 'KMS1'},
            rc=252,
        )

    def test_valid_consistent_args(self):
        self._assert_error_msg(
            emrfs_option_value='SSE=true,RetryCount=5,RetryPeriod=30',
            exception_class_name='InvalidEmrFsArgumentsError',
            error_msg_kwargs={
                'invalid': 'RetryCount and RetryPeriod',
                'parent_object_name': CONSISTENT_OPTION_NAME,
            },
            rc=252,
        )

    def test_valid_cse_kms_args(self):
        self._assert_error_msg(
            emrfs_option_value='Consistent=true,KMSKeyId=k1',
            exception_class_name='InvalidEmrFsArgumentsError',
            error_msg_kwargs={
                'invalid': 'KMSKeyId',
                'parent_object_name': CSE_KMS_OPTION_NAME,
            },
            rc=252,
        )

    def test_valid_cse_custom_args(self):
        self._assert_error_msg(
            emrfs_option_value='Consistent=true,CustomProviderLocation=loc',
            exception_class_name='InvalidEmrFsArgumentsError',
            error_msg_kwargs={
                'invalid': 'CustomProviderLocation',
                'parent_object_name': CSE_CUSTOM_OPTION_NAME,
            },
            rc=252,
        )

    def test_configurations_and_emrfs(self):
        emrfs_option_value = 'Args=[someProperty=someValue]'
        configurations = json.dumps(
            DEFAULT_CONFIGURATIONS, separators=(',', ':')
        )

        cmd = "%s --release-label emr-4.0 --emrfs %s --configurations %s" % (
            DEFAULT_CMD,
            emrfs_option_value,
            configurations,
        )

        expected_emrfs_properties = {'someProperty': 'someValue'}

        emrfs_configuration = copy.deepcopy(EMPTY_EMRFS_CONFIGURATION)
        configurations = copy.deepcopy(DEFAULT_CONFIGURATIONS)
        result = copy.deepcopy(DEFAULT_RESULT)

        emrfs_configuration['Properties'] = expected_emrfs_properties
        configurations.append(emrfs_configuration)
        result['ReleaseLabel'] = 'emr-4.0'
        result['Configurations'] = configurations

        self.assert_params_for_cmd(cmd, result)

    def test_duplicate_emrfs_configuration_exception(self):
        emrfs_option_value = 'Args=[someProperty=someValue]'
        expected_emrfs_properties = {'someProperty': 'someValue'}

        emrfs_configuration = copy.deepcopy(EMPTY_EMRFS_CONFIGURATION)
        emrfs_configuration['Properties'] = expected_emrfs_properties
        configurations = copy.deepcopy(DEFAULT_CONFIGURATIONS)
        configurations.append(EMPTY_EMRFS_CONFIGURATION)

        configurations_json = json.dumps(configurations, separators=(',', ':'))

        cmd = "%s --release-label emr-4.0 --emrfs %s --configurations %s" % (
            DEFAULT_CMD,
            emrfs_option_value,
            configurations_json,
        )

        result = copy.deepcopy(DEFAULT_RESULT)
        result['ReleaseLabel'] = 'emr-4.0'
        result['Configurations'] = configurations

        self.assert_error_msg(cmd, 'DuplicateEmrFsConfigurationError', rc=252)

    def _assert_error_msg(
        self,
        emrfs_option_value,
        exception_class_name,
        error_msg_kwargs,
        rc=255,
    ):
        cmd = "%s --ami-version 3.4 --emrfs %s" % (
            DEFAULT_CMD,
            emrfs_option_value,
        )
        self.assert_error_msg(
            cmd,
            exception_class_name=exception_class_name,
            error_msg_kwargs=error_msg_kwargs,
            rc=rc,
        )

        cmd = "%s --release-label emr-4.0 --emrfs %s" % (
            DEFAULT_CMD,
            emrfs_option_value,
        )
        self.assert_error_msg(
            cmd,
            exception_class_name=exception_class_name,
            error_msg_kwargs=error_msg_kwargs,
            rc=rc,
        )

    def _assert_bootstrap_actions(
        self,
        emrfs_option_value,
        expected_emrfs_ba_key_values,
        expected_emrfs_properties,
        provider_location=None,
    ):
        if expected_emrfs_ba_key_values is not None:
            cmd = "%s --ami-version 3.4 --emrfs %s" % (
                DEFAULT_CMD,
                emrfs_option_value,
            )
            result = copy.deepcopy(DEFAULT_RESULT)
            result['BootstrapActions'] = (
                [self._create_s3_get_ba_config(provider_location)]
                if provider_location is not None
                else []
            )
            result['BootstrapActions'] += [
                self._create_setup_emrfs_ba_config(
                    expected_emrfs_ba_key_values
                )
            ]
            result['AmiVersion'] = '3.4'

            self.assert_params_for_cmd(cmd, result)

        if expected_emrfs_properties is not None:
            cmd = "%s --release-label emr-4.0 --emrfs %s" % (
                DEFAULT_CMD,
                emrfs_option_value,
            )
            result = copy.deepcopy(DEFAULT_RESULT)
            emrfs_configuration = copy.deepcopy(EMPTY_EMRFS_CONFIGURATION)
            emrfs_configuration['Properties'] = expected_emrfs_properties
            result['Configurations'] = [emrfs_configuration]
            result['ReleaseLabel'] = 'emr-4.0'

            self.assert_params_for_cmd(cmd, result)

    def _create_setup_emrfs_ba_config(self, ba_arg_values):
        ba_arg_keys = ['-e' for x in ba_arg_values]
        ba_args = [x for pair in zip(ba_arg_keys, ba_arg_values) for x in pair]

        return {
            'Name': 'Setup EMRFS',
            'ScriptBootstrapAction': {
                'Path': (
                    's3://us-east-1.elasticmapreduce/'
                    'bootstrap-actions/configure-hadoop'
                ),
                'Args': ba_args,
            },
        }

    def _create_s3_get_ba_config(self, provider_location):
        return {
            'Name': 'S3 get',
            'ScriptBootstrapAction': {
                'Path': 'file:/usr/share/aws/emr/scripts/s3get',
                'Args': [
                    '-s',
                    provider_location,
                    '-d',
                    '/usr/share/aws/emr/auxlib',
                    '-f',
                ],
            },
        }



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.autocomplete import parser
from awscli.autoprompt.doc import DocsGetter
from awscli.clidriver import create_clidriver
from awscli.testutils import mock, unittest
from tests.unit.autocomplete import InMemoryIndex


class TestDocsGetter(unittest.TestCase):
    def setUp(self):
        self.driver = create_clidriver()
        self.docs_getter = DocsGetter(self.driver)
        self.index = InMemoryIndex(
            {
                'command_names': {
                    '': [('aws', None)],
                    'aws': [('ec2', None)],
                    'aws.ec2': [('describe-instances', None)],
                },
                'arg_names': {
                    '': {'aws': ['region']},
                    'aws.ec2': {'describe-instances': []},
                },
                'arg_data': {
                    '': {
                        'aws': {
                            'region': (
                                'region',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                        }
                    },
                    'aws.ec2': {'describe-instances': {}},
                },
            }
        )
        self.parser = parser.CLIParser(self.index)

    def test_get_service_command_docs(self):
        parsed_args = self.parser.parse('aws ec2')
        actual_docs = self.docs_getter.get_docs(parsed_args)
        expected_docs = 'Amazon EC2'
        self.assertIn(expected_docs, actual_docs)

    def test_get_service_operation_docs(self):
        parsed_args = self.parser.parse('aws ec2 describe-instances')
        actual_docs = self.docs_getter.get_docs(parsed_args)
        expected_docs = 'Describes the specified instances'
        self.assertIn(expected_docs, actual_docs)

    def test_get_service_command_docs_with_invalid_service_operation(self):
        parsed_args = self.parser.parse('aws ec2 fake')
        actual_docs = self.docs_getter.get_docs(parsed_args)
        expected_docs = 'Amazon EC2'
        self.assertIn(expected_docs, actual_docs)

    def test_get_top_level_aws_docs_if_no_command_specified(self):
        parsed_args = self.parser.parse('aws')
        actual_docs = self.docs_getter.get_docs(parsed_args)
        expected_docs = 'The AWS Command Line Interface'
        self.assertIn(expected_docs, actual_docs)

    def test_get_top_level_aws_docs_if_service_command_is_invalid(self):
        parsed_args = self.parser.parse('aws fake')
        actual_docs = self.docs_getter.get_docs(parsed_args)
        expected_docs = 'The AWS Command Line Interface'
        self.assertIn(expected_docs, actual_docs)

    def test_get_service_operation_docs_if_input_is_vanild(self):
        parsed_args = self.parser.parse('aws ec2 describe-instances fake')
        actual_docs = self.docs_getter.get_docs(parsed_args)
        expected_docs = 'Describes the specified instances'
        self.assertIn(expected_docs, actual_docs)

    def test_cache_is_used(self):
        parsed_args = self.parser.parse('aws ec2 describe-instances fake')
        self.docs_getter.get_doc_content = mock.Mock()
        self.docs_getter.get_docs(parsed_args)
        self.assertEqual(self.docs_getter.get_doc_content.call_count, 1)
        self.docs_getter.get_docs(parsed_args)
        self.assertEqual(self.docs_getter.get_doc_content.call_count, 1)



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from argparse import Namespace

from awscli.customizations.codedeploy.deregister import Deregister
from awscli.customizations.exceptions import (
    ConfigurationError,
    ParamValidationError,
)
from awscli.testutils import mock, unittest


class TestDeregister(unittest.TestCase):
    def setUp(self):
        self.instance_name = 'instance-name'
        self.tags = [{'Key': 'k1', 'Value': 'v1'}]
        self.iam_user_arn = (
            f'arn:aws:iam::012345678912:user/{self.instance_name}'
        )
        self.access_key_id = 'ACCESSKEYID'
        self.region = 'us-east-1'
        self.policy_name = 'codedeploy-agent'
        self.endpoint_url = 'https://codedeploy.aws.amazon.com'

        self.args = Namespace()
        self.args.instance_name = self.instance_name
        self.args.no_delete_iam_user = False

        self.globals = Namespace()
        self.globals.region = self.region
        self.globals.endpoint_url = self.endpoint_url
        self.globals.verify_ssl = False

        self.codedeploy = mock.MagicMock()
        self.codedeploy.get_on_premises_instance.return_value = {
            'instanceInfo': {'iamUserArn': self.iam_user_arn, 'tags': None}
        }

        self.iam = mock.MagicMock()
        self.list_user_policies = mock.MagicMock()
        self.list_user_policies.paginate.return_value = [
            {'PolicyNames': [self.policy_name]}
        ]
        self.list_access_keys = mock.MagicMock()
        self.list_access_keys.paginate.return_value = [
            {'AccessKeyMetadata': [{'AccessKeyId': self.access_key_id}]}
        ]
        self.iam.get_paginator.side_effect = [
            self.list_user_policies,
            self.list_access_keys,
        ]

        self.session = mock.MagicMock()
        self.session.create_client.side_effect = [self.codedeploy, self.iam]
        self.deregister = Deregister(self.session)

    def test_deregister_throws_on_invalid_region(self):
        self.globals.region = None
        self.session.get_config_variable.return_value = None
        error_msg = 'Region not specified.'
        with self.assertRaisesRegex(ConfigurationError, error_msg):
            self.deregister._run_main(self.args, self.globals)

    def test_deregister_throws_on_invalid_instance_name(self):
        self.args.instance_name = 'invalid%@^&%#&'
        with self.assertRaisesRegex(
            ParamValidationError, 'Instance name contains invalid characters.'
        ):
            self.deregister._run_main(self.args, self.globals)

    def test_deregister_creates_clients(self):
        self.deregister._run_main(self.args, self.globals)
        self.session.create_client.assert_has_calls(
            [
                mock.call(
                    'codedeploy',
                    region_name=self.region,
                    endpoint_url=self.endpoint_url,
                    verify=self.globals.verify_ssl,
                ),
                mock.call('iam', region_name=self.region),
            ]
        )

    def test_deregister_with_tags(self):
        self.codedeploy.get_on_premises_instance.return_value = {
            'instanceInfo': {
                'iamUserArn': self.iam_user_arn,
                'tags': self.tags,
            }
        }
        self.deregister._run_main(self.args, self.globals)
        self.codedeploy.get_on_premises_instance.assert_called_with(
            instanceName=self.instance_name
        )
        self.assertIn('iam_user_arn', self.args)
        self.assertEqual(self.iam_user_arn, self.args.iam_user_arn)
        self.assertIn('user_name', self.args)
        self.assertEqual(self.instance_name, self.args.user_name)
        self.assertIn('tags', self.args)
        self.assertEqual(self.tags, self.args.tags)
        self.codedeploy.remove_tags_from_on_premises_instances.assert_called_with(
            tags=self.tags, instanceNames=[self.instance_name]
        )
        self.codedeploy.deregister_on_premises_instance.assert_called_with(
            instanceName=self.instance_name
        )

    def test_deregister_with_no_tags(self):
        self.codedeploy.get_on_premises_instance.return_value = {
            'instanceInfo': {'iamUserArn': self.iam_user_arn, 'tags': None}
        }
        self.deregister._run_main(self.args, self.globals)
        self.codedeploy.get_on_premises_instance.assert_called_with(
            instanceName=self.instance_name
        )
        self.assertIn('iam_user_arn', self.args)
        self.assertEqual(self.iam_user_arn, self.args.iam_user_arn)
        self.assertIn('user_name', self.args)
        self.assertEqual(self.instance_name, self.args.user_name)
        self.assertIn('tags', self.args)
        self.assertEqual(None, self.args.tags)
        self.assertFalse(
            self.codedeploy.remove_tags_from_on_premises_instances.called
        )
        self.codedeploy.deregister_on_premises_instance.assert_called_with(
            instanceName=self.instance_name
        )

    def test_deregister_with_delete_iam_user(self):
        self.args.no_delete_iam_user = False
        self.deregister._run_main(self.args, self.globals)
        self.codedeploy.get_on_premises_instance.assert_called_with(
            instanceName=self.instance_name
        )
        self.codedeploy.deregister_on_premises_instance.assert_called_with(
            instanceName=self.instance_name
        )
        self.iam.get_paginator.assert_has_calls(
            [mock.call('list_user_policies'), mock.call('list_access_keys')]
        )
        self.list_user_policies.paginate.assert_called_with(
            UserName=self.instance_name
        )
        self.iam.delete_user_policy.assert_called_with(
            UserName=self.instance_name, PolicyName=self.policy_name
        )
        self.list_access_keys.paginate.assert_called_with(
            UserName=self.instance_name
        )
        self.iam.delete_access_key.assert_called_with(
            UserName=self.instance_name, AccessKeyId=self.access_key_id
        )
        self.iam.delete_user.assert_called_with(UserName=self.instance_name)

    def test_deregister_with_no_delete_iam_user(self):
        self.args.no_delete_iam_user = True
        self.deregister._run_main(self.args, self.globals)
        self.codedeploy.get_on_premises_instance.assert_called_with(
            instanceName=self.instance_name
        )
        self.codedeploy.deregister_on_premises_instance.assert_called_with(
            instanceName=self.instance_name
        )
        self.assertFalse(self.iam.get_paginator.called)
        self.assertFalse(self.list_user_policies.paginate.called)
        self.assertFalse(self.iam.delete_user_policy.called)
        self.assertFalse(self.list_access_keys.paginate.called)
        self.assertFalse(self.iam.delete_access_key.called)
        self.assertFalse(self.iam.delete_user.called)


if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.emr import exceptions, sshutils
from awscli.testutils import mock, unittest


class TestSSHUtils(unittest.TestCase):
    @mock.patch('awscli.customizations.emr.sshutils.emrutils')
    def test_validate_and_find_master_dns_waits(self, emrutils):
        emrutils.get_cluster_state.return_value = 'STARTING'
        session = mock.Mock()
        client = mock.Mock()
        emrutils.get_client.return_value = client

        sshutils.validate_and_find_master_dns(session, None, 'cluster-id')

        # We should have:
        # 1. Waiter for the cluster to be running.
        client.get_waiter.assert_called_with('cluster_running')
        client.get_waiter.return_value.wait.assert_called_with(
            ClusterId='cluster-id'
        )

        # 2. Found the master public DNS
        self.assertTrue(emrutils.find_master_dns.called)

    @mock.patch('awscli.customizations.emr.sshutils.emrutils')
    def test_cluster_in_terminated_states(self, emrutils):
        emrutils.get_cluster_state.return_value = 'TERMINATED'
        with self.assertRaises(exceptions.ClusterTerminatedError):
            sshutils.validate_and_find_master_dns(
                mock.Mock(), None, 'cluster-id'
            )

    @mock.patch('awscli.customizations.emr.sshutils.emrutils')
    def test_ssh_scp_key_file_format(self, emrutils):
        def which_side_effect(program):
            if program == 'ssh' or program == 'scp':
                return '/some/path'

        emrutils.which.side_effect = which_side_effect

        key_file1 = 'key.abc'
        sshutils.validate_ssh_with_key_file(key_file1)
        sshutils.validate_scp_with_key_file(key_file1)

        key_file2 = 'key'
        sshutils.validate_ssh_with_key_file(key_file2)
        sshutils.validate_scp_with_key_file(key_file2)



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import argparse
import datetime

from botocore.session import Session

from awscli.compat import StringIO, ensure_text_type
from awscli.customizations.history.db import DatabaseRecordReader
from awscli.customizations.history.list import (
    ListCommand,
    RecordAdapter,
    TextFormatter,
)
from awscli.testutils import mock, unittest
from awscli.utils import OutputStreamFactory


class TestRecordAdapter(unittest.TestCase):
    def test_no_records_no_next(self):
        source = iter([])
        adapter = RecordAdapter(source)
        self.assertFalse(adapter.has_next())

    def test_one_record_has_next(self):
        source = iter([1])
        adapter = RecordAdapter(source)
        self.assertTrue(adapter.has_next())

    def test_can_iterate(self):
        source = iter([1, 2, 3])
        adapter = RecordAdapter(source)
        result = [r for r in adapter]
        self.assertEqual(result, [1, 2, 3])


class TestTextFormatter(unittest.TestCase):
    _COL_WIDTHS = {'id_a': 10, 'timestamp': 23, 'args': 10, 'rc': 10}

    def setUp(self):
        self.output_stream = StringIO()
        self.formatter = TextFormatter(self._COL_WIDTHS, self.output_stream)

        self.timestamp = 1511376242067
        command_time = datetime.datetime.fromtimestamp(self.timestamp / 1000)
        self.formatted_time = datetime.datetime.strftime(
            command_time, '%Y-%m-%d %I:%M:%S %p'
        )

    def _format_records(self, records):
        adapter = RecordAdapter(iter(records))
        self.formatter(adapter)

    def test_can_emit_single_row(self):
        self._format_records(
            [
                {
                    'id_a': 'foo',
                    'timestamp': self.timestamp,
                    'args': '["s3", "ls"]',
                    'rc': 0,
                }
            ]
        )
        expected_output = 'foo       %s s3 ls     0\n' % self.formatted_time
        actual_output = ensure_text_type(self.output_stream.getvalue())
        self.assertEqual(expected_output, actual_output)

    def test_can_emit_multiple_rows(self):
        self._format_records(
            [
                {
                    'id_a': 'foo',
                    'timestamp': self.timestamp,
                    'args': '["s3", "ls"]',
                    'rc': 0,
                },
                {
                    'id_a': 'bar',
                    'timestamp': self.timestamp,
                    'args': '["s3", "cp"]',
                    'rc': 1,
                },
            ]
        )
        expected_output = (
            'foo       %s s3 ls     0\n' 'bar       %s s3 cp     1\n'
        ) % (self.formatted_time, self.formatted_time)
        actual_output = ensure_text_type(self.output_stream.getvalue())
        self.assertEqual(expected_output, actual_output)

    def test_can_truncate_args(self):
        # Truncate the argument if it won't fit in the space allotted to the
        # arguments field.
        self._format_records(
            [
                {
                    'id_a': 'foo',
                    'timestamp': self.timestamp,
                    'args': (
                        '["s3", "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
                        'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"]'
                    ),
                    'rc': 0,
                }
            ]
        )
        expected_output = 'foo       %s s3 aaa... 0\n' % self.formatted_time
        actual_output = ensure_text_type(self.output_stream.getvalue())
        self.assertEqual(expected_output, actual_output)


class TestListCommand(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock(Session)

        self.output_stream_factory = mock.Mock(OutputStreamFactory)

        # MagicMock is needed because it can handle context managers.
        # Normal Mock will throw AttributeErrors
        output_stream_context = mock.MagicMock()
        self.output_stream = mock.Mock()
        output_stream_context.__enter__.return_value = self.output_stream

        self.output_stream_factory.get_output_stream.return_value = (
            output_stream_context
        )

        self.db_reader = mock.Mock(DatabaseRecordReader)
        self.db_reader.iter_all_records.return_value = iter([])

        self.list_cmd = ListCommand(
            self.session, self.db_reader, self.output_stream_factory
        )

        self.parsed_args = argparse.Namespace()

        self.parsed_globals = argparse.Namespace()
        self.parsed_globals.color = 'auto'

    def _make_record(self, cid, time, args, rc):
        record = {'id_a': cid, 'timestamp': time, 'args': args, 'rc': rc}
        return record

    def test_does_call_iter_all_records(self):
        with self.assertRaises(RuntimeError):
            self.list_cmd._run_main(self.parsed_args, self.parsed_globals)
        self.assertTrue(self.db_reader.iter_all_records.called)

    def test_list_does_write_values_to_stream(self):
        self.db_reader.iter_all_records.return_value = iter(
            [self._make_record('abc', 1511376242067, '["s3", "ls"]', '0')]
        )
        self.list_cmd._run_main(self.parsed_args, self.parsed_globals)
        self.assertTrue(self.output_stream.write.called)

    @mock.patch('awscli.customizations.history.list.OutputStreamFactory')
    def test_uses_sr_less_flags(self, mock_factory):
        ListCommand(self.session)
        mock_factory.assert_called_with(self.session, default_less_flags='SR')



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import collections
import itertools
import json
import logging
import re

from botocore.utils import is_valid_endpoint_url
from prompt_toolkit import prompt as ptk_prompt
from prompt_toolkit.application import get_app
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import ValidationError, Validator

from awscli.customizations.sso.utils import (
    parse_sso_registration_scopes,
)

logger = logging.getLogger(__name__)


class ValidatorWithDefault(Validator):
    def __init__(self, default=None):
        super().__init__()
        self._default = default

    def _raise_validation_error(self, document, message):
        index = len(document.text)
        raise ValidationError(index, message)


class StartUrlValidator(ValidatorWithDefault):
    def validate(self, document):
        # If there's a default, allow an empty prompt
        if not document.text and self._default:
            return
        if not is_valid_endpoint_url(document.text):
            self._raise_validation_error(document, 'Not a valid Start URL')


class RequiredInputValidator(ValidatorWithDefault):
    def validate(self, document):
        if document.text or self._default:
            return
        self._raise_validation_error(document, 'A value is required')


class ScopesValidator(ValidatorWithDefault):
    def validate(self, document):
        # If there's a default, allow an empty prompt
        if not document.text and self._default:
            return
        if not self._is_comma_separated_list(document.text):
            self._raise_validation_error(
                document, 'Scope values must be separated by commas'
            )

    def _is_comma_separated_list(self, value):
        scopes = value.split(',')
        for scope in scopes:
            if re.findall(r'\s', scope.strip()):
                return False
        return True


class PTKPrompt:
    _DEFAULT_PROMPT_FORMAT = '{prompt_text} [{current_value}]: '

    def __init__(self, prompter=None):
        if prompter is None:
            prompter = ptk_prompt
        self._prompter = prompter

    def _create_completer(self, completions):
        if completions is None:
            completions = []
        completer_kwargs = {
            'words': completions,
            'pattern': re.compile(r'\S+'),
        }
        if isinstance(completions, dict):
            completer_kwargs['meta_dict'] = completions
            completer_kwargs['words'] = list(completions.keys())
        return WordCompleter(**completer_kwargs)

    def get_value(
        self,
        current_value,
        prompt_text='',
        completions=None,
        validator=None,
        toolbar=None,
        prompt_fmt=None,
    ):
        if prompt_fmt is None:
            prompt_fmt = self._DEFAULT_PROMPT_FORMAT
        prompt_string = prompt_fmt.format(
            prompt_text=prompt_text, current_value=current_value
        )
        prompter_kwargs = {
            'validator': validator,
            'validate_while_typing': False,
            'completer': self._create_completer(completions),
            'complete_while_typing': True,
            'style': self._get_prompt_style(),
        }
        if toolbar:
            prompter_kwargs['bottom_toolbar'] = toolbar
            prompter_kwargs['refresh_interval'] = 0.2
        response = self._prompter(prompt_string, **prompter_kwargs)
        # Strip any extra white space
        response = response.strip()
        if not response:
            # If the user hits enter, we return the current/default value
            response = current_value
        return response

    def _get_prompt_style(self):
        return Style.from_dict(
            {
                'bottom-toolbar': 'noreverse',
            }
        )


class SSOSessionConfigurationPrompter:
    _DEFAULT_SSO_SCOPE = 'sso:account:access'
    _KNOWN_SSO_SCOPES = {
        'sso:account:access': (
            'Grants access to AWS IAM Identity Center accounts and permission '
            'sets'
        )
    }

    def __init__(self, botocore_session, prompter):
        self._botocore_session = botocore_session
        self._prompter = prompter
        self._sso_sessions = self._botocore_session.full_config.get(
            'sso_sessions', {}
        )
        self._sso_session = None
        self.sso_session_config = {}

    @property
    def sso_session(self):
        return self._sso_session

    @sso_session.setter
    def sso_session(self, value):
        self._sso_session = value
        self.sso_session_config = self._sso_sessions.get(
            self._sso_session, {}
        ).copy()

    def prompt_for_sso_session(self, required=True):
        prompt_text = 'SSO session name'
        prompt_fmt = None
        validator_cls = None
        if required:
            validator_cls = RequiredInputValidator
        if not self.sso_session:
            prompt_fmt = f'{prompt_text}: '
            if not required:
                prompt_fmt = f'{prompt_text} (Recommended): '
        sso_session = self._prompt_for(
            'sso_session',
            prompt_text,
            completions=sorted(self._sso_sessions),
            toolbar=self._get_sso_session_toolbar,
            validator_cls=validator_cls,
            prompt_fmt=prompt_fmt,
            current_value=self.sso_session,
        )
        self.sso_session = sso_session
        return sso_session

    def prompt_for_sso_start_url(self):
        return self._prompt_for(
            'sso_start_url',
            'SSO start URL',
            completions=self._get_potential_start_urls(),
            validator_cls=StartUrlValidator,
        )

    def prompt_for_sso_region(self):
        return self._prompt_for(
            'sso_region',
            'SSO region',
            completions=self._get_potential_sso_regions(),
            validator_cls=RequiredInputValidator,
        )

    def prompt_for_sso_registration_scopes(self):
        if 'sso_registration_scopes' not in self.sso_session_config:
            self.sso_session_config['sso_registration_scopes'] = (
                self._DEFAULT_SSO_SCOPE
            )
        raw_scopes = self._prompt_for(
            'sso_registration_scopes',
            'SSO registration scopes',
            completions=self._get_potential_sso_registrations_scopes(),
            validator_cls=ScopesValidator,
        )
        return parse_sso_registration_scopes(raw_scopes)

    def _prompt_for(
        self,
        config_name,
        text,
        completions=None,
        validator_cls=None,
        toolbar=None,
        prompt_fmt=None,
        current_value=None,
    ):
        if current_value is None:
            current_value = self.sso_session_config.get(config_name)
        validator = None
        if validator_cls:
            validator = validator_cls(current_value)
        value = self._prompter.get_value(
            current_value,
            text,
            completions=completions,
            validator=validator,
            toolbar=toolbar,
            prompt_fmt=prompt_fmt,
        )
        if value:
            self.sso_session_config[config_name] = value
        return value

    def _get_sso_session_toolbar(self):
        current_input = get_app().current_buffer.document.text
        if current_input in self._sso_sessions:
            selected_sso_config = self._sso_sessions[current_input]
            return FormattedText(
                [
                    ('', self._get_toolbar_border()),
                    ('', '\n'),
                    (
                        'bold',
                        f'Configuration for SSO session: {current_input}\n\n',
                    ),
                    ('', json.dumps(selected_sso_config, indent=2)),
                ]
            )

    def _get_toolbar_border(self):
        horizontal_line_char = '\u2500'
        return horizontal_line_char * get_app().output.get_size().columns

    def _get_potential_start_urls(self):
        profiles = self._botocore_session.full_config.get('profiles', {})
        configs_to_search = itertools.chain(
            profiles.values(), self._sso_sessions.values()
        )
        potential_start_urls = set()
        for config_to_search in configs_to_search:
            if 'sso_start_url' in config_to_search:
                start_url = config_to_search['sso_start_url']
                potential_start_urls.add(start_url)
        return list(potential_start_urls)

    def _get_potential_sso_regions(self):
        return self._botocore_session.get_available_regions('sso-oidc')

    def _get_potential_sso_registrations_scopes(self):
        potential_scopes = self._KNOWN_SSO_SCOPES.copy()
        scopes_to_sessions = self._get_previously_used_scopes_to_sso_sessions()
        for scope, sso_sessions in scopes_to_sessions.items():
            if scope not in potential_scopes:
                potential_scopes[scope] = (
                    f'Used in SSO sessions: {", ".join(sso_sessions)}'
                )
        return potential_scopes

    def _get_previously_used_scopes_to_sso_sessions(self):
        scopes_to_sessions = collections.defaultdict(list)
        for sso_session, sso_session_config in self._sso_sessions.items():
            if 'sso_registration_scopes' in sso_session_config:
                parsed_scopes = parse_sso_registration_scopes(
                    sso_session_config['sso_registration_scopes']
                )
                for parsed_scope in parsed_scopes:
                    scopes_to_sessions[parsed_scope].append(sso_session)
        return scopes_to_sessions



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from copy import deepcopy

from awscli.testutils import mock
from tests.unit.customizations.emr import (
    EMRBaseAWSCommandParamsTest as BaseAWSCommandParamsTest,
)


class TestDisableHBaseBackups(BaseAWSCommandParamsTest):
    prefix = 'emr disable-hbase-backups'
    DISABLE_FULL_BACKUP = '--disable-full-backups'
    DISABLE_INCR_BACKUP = '--disable-incremental-backups'
    default_steps = [
        {
            'HadoopJarStep': {
                'Args': [
                    'emr.hbase.backup.Main',
                    '--set-scheduled-backup',
                    'false',
                ],
                'Jar': '/home/hadoop/lib/hbase.jar',
            },
            'Name': 'Modify Backup Schedule',
            'ActionOnFailure': 'CANCEL_AND_WAIT',
        }
    ]

    def test_disable_hbase_backups_full(self):
        args = ' --cluster-id j-ABCD --full'
        cmdline = self.prefix + args

        steps = deepcopy(self.default_steps)
        steps[0]['HadoopJarStep']['Args'].append(self.DISABLE_FULL_BACKUP)
        result = {'JobFlowId': 'j-ABCD', 'Steps': steps}

        self.assert_params_for_cmd(cmdline, result)

    def test_disable_hbase_backups_incremental(self):
        args = ' --cluster-id j-ABCD --incremental'
        cmdline = self.prefix + args

        steps = deepcopy(self.default_steps)
        steps[0]['HadoopJarStep']['Args'].append(self.DISABLE_INCR_BACKUP)
        result = {'JobFlowId': 'j-ABCD', 'Steps': steps}

        self.assert_params_for_cmd(cmdline, result)

    def test_disable_hbase_backups_both(self):
        args = ' --cluster-id j-ABCD --full --incremental'
        cmdline = self.prefix + args

        steps = deepcopy(self.default_steps)
        steps[0]['HadoopJarStep']['Args'].append(self.DISABLE_FULL_BACKUP)
        steps[0]['HadoopJarStep']['Args'].append(self.DISABLE_INCR_BACKUP)
        result = {'JobFlowId': 'j-ABCD', 'Steps': steps}

        self.assert_params_for_cmd(cmdline, result)

    def test_disable_hbase_backups_none(self):
        args = ' --cluster-id j-ABCD'
        cmdline = self.prefix + args
        expected_error_msg = (
            '\naws: [ERROR]: An error occurred (ParamValidation): Should specify at least one of --full' + ' and --incremental.\n'
        )
        result = self.run_cmd(cmdline, 252)

        self.assertEqual(expected_error_msg, result[1])

    @mock.patch('awscli.customizations.emr.' 'emrutils.get_release_label')
    def test_unsupported_command_on_release_based_cluster_error(
        self, grl_patch
    ):
        grl_patch.return_value = 'emr-4.0'
        args = ' --cluster-id j-ABCD --full'
        cmdline = self.prefix + args
        expected_error_msg = (
            "\naws: [ERROR]: An error occurred (ParamValidation): disable-hbase-backups"
            " is not supported with 'emr-4.0' release.\n"
        )
        result = self.run_cmd(cmdline, 252)

        self.assertEqual(result[1], expected_error_msg)


if __name__ == "__main__":
    unittest.main()



#
# aws-cli documentation build configuration file, created by
# sphinx-quickstart on Fri Feb  1 12:57:38 2013.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime
import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('.'))

import bootstrapdocs

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'notfound.extension',
    'crosslinker',
]
notfound_context = {
    'title': 'Page not found',
    'body': '<h1>Page not found</h1>\n\n'
    'Sorry, the page you requested could not be found.',
}
notfound_pagename = '_404'
# notfound.extension changes all the relative links to links like
# "/en/latest/_static/**" and we use "notfound_urls_prefix" key
# to set our path prefix
notfound_urls_prefix = '/v2/documentation/api/latest/'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'AWS CLI Command Reference'
copyright = f'{datetime.datetime.now().year}, Amazon Web Services'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '2.0'
# The full version, including alpha/beta/rc tags.
release = '2.34.56'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:

# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['examples']

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'guzzle_sphinx_theme.GuzzleStyle'
# pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = 'pyramid'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {
#     'bodyfont': '"Andale Mono", Courier, monospace;'
#     }

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = ['.']

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = "AWS CLI %s Command Reference" % release

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    '**': [
        'sidebarlogo.html',
        'localtoc.html',
        'searchbox.html',
        'feedback.html',
        'userguide.html',
    ]
}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'aws-clidoc'

# Adds an HTML table visitor to apply Bootstrap table classes
html_translator_class = 'guzzle_sphinx_theme.HTMLTranslator'
p = os.path.abspath('.')
p = os.path.join(p, 'guzzle_sphinx_theme')
html_theme_path = [p]
html_theme = 'guzzle_sphinx_theme'
# Register the theme as an extension to generate a sitemap.xml
extensions.append("guzzle_sphinx_theme")

html_theme_options = {
    # Set the name of the project to appear in the nav menu
    "project_nav_name": "AWS CLI Command Reference",
    # Set your GitHub user and repo to enable GitHub stars links
    "github_user": "aws",
    "github_repo": "aws-cli",
    # Set to true to bind left and right key events to turn the page
    "bind_key_events": False,
    # Specify a base_url used to generate sitemap.xml links. If not
    # specified, then no sitemap will be built.
    "base_url": "http://docs.aws.amazon.com/cli/latest/",
}


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        'index',
        'aws-cli.tex',
        'AWS CLI Documentation',
        'Amazon Web Services',
        'manual',
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).

man_pages = [
    ('reference/index', 'aws', 'The main command', '', 1),
    (
        'reference/autoscaling/index',
        'aws-autoscaling',
        'The autoscaling service',
        '',
        1,
    ),
    (
        'reference/cloudformation/index',
        'aws-cloudformation',
        'AWS CloudFormation',
        '',
        1,
    ),
    (
        'reference/cloudwatch/index',
        'aws-cloudwatch',
        'Amazon CloudWatch',
        '',
        1,
    ),
    (
        'reference/datapipeline/index',
        'aws-datapipeline',
        'AWS Data Pipeline',
        '',
        1,
    ),
    (
        'reference/directconnect/index',
        'aws-directconnect',
        'AWS Direct Connect',
        '',
        1,
    ),
    ('reference/dynamodb/index', 'aws-dynamodb', 'Amazon DynamoDB', '', 1),
    ('reference/ec2/index', 'aws-ec2', 'Amazon Elastic Compute Cloud', '', 1),
    (
        'reference/elasticache/index',
        'aws-elasticache',
        'Amazon ElastiCache',
        '',
        1,
    ),
    (
        'reference/elasticbeanstalk/index',
        'aws-elasticbeanstalk',
        'AWS Elastic Beanstalk',
        '',
        1,
    ),
    ('reference/elb/index', 'aws-elb', 'Elastic Load Balancing', '', 1),
    ('reference/emr/index', 'aws-emr', 'Amazon Elastic MapReduce', '', 1),
    (
        'reference/iam/index',
        'aws-iam',
        'AWS Identity and Access Management',
        '',
        1,
    ),
    (
        'reference/importexport/index',
        'aws-importexport',
        'AWS Import/Export',
        '',
        1,
    ),
    (
        'reference/rds/index',
        'aws-rds',
        'Amazon Relational Database Service',
        '',
        1,
    ),
    ('reference/redshift/index', 'aws-redshift', 'Amazon Redshift', '', 1),
    ('reference/route53/index', 'aws-route53', 'Amazon Route 53', '', 1),
    ('reference/s3/index', 'aws-s3', 'Amazon Simple Storage Service', '', 1),
    ('reference/ses/index', 'aws-ses', 'Amazon Simple Email Service', '', 1),
    (
        'reference/sns/index',
        'aws-sns',
        'Amazon Simple Notification Service',
        '',
        1,
    ),
    ('reference/sqs/index', 'aws-sqs', 'Amazon Simple Queue Service', '', 1),
    (
        'reference/storagegateway/index',
        'aws-storagegateway',
        'AWS Storage Gateway',
        '',
        1,
    ),
    ('reference/sts/index', 'aws-sts', 'AWS Security Token Service', '', 1),
    ('reference/support/index', 'aws-support', 'AWS Support', '', 1),
    (
        'reference/swf/index',
        'aws-swf',
        'Amazon Simple Workflow Service',
        '',
        1,
    ),
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = []


# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from decimal import Decimal

import pytest

from awscli.customizations.dynamodb.exceptions import (
    EmptyExpressionError,
    LexerError,
)
from awscli.customizations.dynamodb.lexer import Lexer

STRING_TOKENS = ['and', 'between', 'in', 'or', 'not']


@pytest.mark.parametrize(
    'case',
    [
        ('foo', [{'type': 'unquoted_identifier', 'value': 'foo'}]),
        ("'foo'", [{'type': 'identifier', 'value': 'foo'}]),
        ("'f\\'oo'", [{'type': 'identifier', 'value': "f'oo"}]),
        ('"spam"', [{'type': 'literal', 'value': 'spam'}]),
        ('"s\\"pam"', [{'type': 'literal', 'value': 's"pam'}]),
        ('100', [{'type': 'literal', 'value': Decimal('100')}]),
        ('-100', [{'type': 'literal', 'value': Decimal('-100')}]),
        ('1.01', [{'type': 'literal', 'value': Decimal('1.01')}]),
        ('1.01e6', [{'type': 'literal', 'value': Decimal('1.01e6')}]),
        ('1.01E6', [{'type': 'literal', 'value': Decimal('1.01e6')}]),
        ('1.01e+6', [{'type': 'literal', 'value': Decimal('1.01e6')}]),
        ('1.01e-6', [{'type': 'literal', 'value': Decimal('1.01e-6')}]),
        (
            "foo, 'bar', \"baz\"",
            [
                {'type': 'unquoted_identifier', 'value': 'foo'},
                {'type': 'comma', 'value': ','},
                {'type': 'identifier', 'value': 'bar'},
                {'type': 'comma', 'value': ','},
                {'type': 'literal', 'value': 'baz'},
            ],
        ),
        ('b"4pyT"', [{'type': 'literal', 'value': b'\xe2\x9c\x93'}]),
        ('boo', [{'type': 'unquoted_identifier', 'value': 'boo'}]),
        (
            'b[0]',
            [
                {'type': 'unquoted_identifier', 'value': 'b'},
                {'type': 'lbracket', 'value': '['},
                {'type': 'literal', 'value': Decimal('0')},
                {'type': 'rbracket', 'value': ']'},
            ],
        ),
        ('true', [{'type': 'literal', 'value': True}]),
        ('false', [{'type': 'literal', 'value': False}]),
        ('null', [{'type': 'literal', 'value': None}]),
    ],
)
def test_lexer(case):
    LexTester().assert_tokens(case[0], case[1])


@pytest.mark.parametrize(
    "token,expected_token_type",
    [
        ('.', 'dot'),
        (',', 'comma'),
        (':', 'colon'),
        ('(', 'lparen'),
        (')', 'rparen'),
        ('{', 'lbrace'),
        ('}', 'rbrace'),
        ('[', 'lbracket'),
        (']', 'rbracket'),
        ('=', 'eq'),
        ('>', 'gt'),
        ('>=', 'gte'),
        ('<', 'lt'),
        ('<=', 'lte'),
        ('<>', 'ne'),
    ],
)
def test_simple_tokens(token, expected_token_type):
    LexTester().assert_tokens(
        token, [{'type': expected_token_type, 'value': token}]
    )


@pytest.mark.parametrize("token", STRING_TOKENS)
def test_string_token(token):
    LexTester().assert_tokens(token, [{'type': token, 'value': token}])


@pytest.mark.parametrize("token", STRING_TOKENS)
def test_string_token_upper(token):
    LexTester().assert_tokens(
        token.upper(), [{'type': token, 'value': token.upper()}]
    )


@pytest.mark.parametrize("token", STRING_TOKENS)
def test_string_token_capitalize(token):
    LexTester().assert_tokens(
        token.capitalize(), [{'type': token, 'value': token.capitalize()}]
    )


def test_lexer_empty_error():
    LexTester().assert_empty_error('')


@pytest.mark.parametrize(
    "expression,error_part",
    [
        ("'", "'\n^"),
        ('"', '"\n^'),
        ('-', '-\n^'),
        ('1e-', '1e-\n  ^'),
        ('1ex', '1ex\n  ^'),
        ('1.', '1.\n ^'),
        ('1.x', '1.x\n  ^'),
        ('&', '&\n^'),
        ('|', '|\n^'),
        ('b"', 'b"\n ^'),
        ('b"&"', 'b"&"\n^'),
        # Invalid padding
        ('b"898989;;"', 'b"898989;;"\n^'),
    ],
)
def test_lexer_error(expression, error_part):
    LexTester().assert_lex_error(expression, error_part)


class LexTester:
    def __init__(self):
        self.lexer = Lexer()

    def assert_tokens(self, expression, expected):
        actual = self.lexer.tokenize(expression)
        simple_tokens = [
            {'type': t['type'], 'value': t['value']} for t in actual
        ]
        assert simple_tokens.pop()['type'] == 'eof'
        assert simple_tokens == expected

    def assert_lex_error(self, expression, error_part):
        try:
            list(self.lexer.tokenize(expression))
            raise AssertionError(
                'LexerError not raised for expression: %s' % expression
            )
        except LexerError as e:
            assert error_part in str(e)

    def assert_empty_error(self, expression):
        with pytest.raises(EmptyExpressionError):
            list(self.lexer.tokenize(expression))



#!/usr/bin/env python
# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestCreateCacheCluster(BaseAWSCommandParamsTest):
    maxDiff = None

    prefix = 'elasticache create-cache-cluster '

    def test_create_cache_cluster(self):
        args = (
            '--cache-cluster-id cachecluster-us-east-1c '
            '--num-cache-nodes 1 '
            '--cache-node-type cache.m1.small '
            '--engine memcached '
            '--engine-version 1.4.5 '
            '--cache-security-group-names group1 group2 '
            '--preferred-availability-zone us-east-1c '
            '--auto-minor-version-upgrade '
            '--preferred-maintenance-window fri:08:00-fri:09:00'
        )
        cmdline = self.prefix + args
        result = {
            'AutoMinorVersionUpgrade': True,
            'CacheClusterId': 'cachecluster-us-east-1c',
            'CacheNodeType': 'cache.m1.small',
            'CacheSecurityGroupNames': ['group1', 'group2'],
            'Engine': 'memcached',
            'EngineVersion': '1.4.5',
            'NumCacheNodes': 1,
            'PreferredAvailabilityZone': 'us-east-1c',
            'PreferredMaintenanceWindow': 'fri:08:00-fri:09:00',
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_create_cache_cluster_no_auto_minor_upgrade(self):
        args = (
            '--cache-cluster-id cachecluster-us-east-1c '
            '--num-cache-nodes 1 '
            '--cache-node-type cache.m1.small '
            '--engine memcached '
            '--engine-version 1.4.5 '
            '--cache-security-group-names group1 group2 '
            '--preferred-availability-zone us-east-1c '
            '--no-auto-minor-version-upgrade '
            '--preferred-maintenance-window fri:08:00-fri:09:00'
        )
        cmdline = self.prefix + args
        result = {
            'AutoMinorVersionUpgrade': False,
            'CacheClusterId': 'cachecluster-us-east-1c',
            'CacheNodeType': 'cache.m1.small',
            'CacheSecurityGroupNames': ['group1', 'group2'],
            'Engine': 'memcached',
            'EngineVersion': '1.4.5',
            'NumCacheNodes': 1,
            'PreferredAvailabilityZone': 'us-east-1c',
            'PreferredMaintenanceWindow': 'fri:08:00-fri:09:00',
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_minor_upgrade_arg_not_specified(self):
        args = (
            '--cache-cluster-id cachecluster-us-east-1c '
            '--num-cache-nodes 1 '
            '--cache-node-type cache.m1.small '
            '--engine memcached '
            '--engine-version 1.4.5 '
            '--cache-security-group-names group1 group2 '
            '--preferred-availability-zone us-east-1c '
            '--preferred-maintenance-window fri:08:00-fri:09:00'
        )
        cmdline = self.prefix + args
        # Note how if neither '--auto-minor-version-upgrade' nor
        # '--no-auto-minor-version-upgrade' is specified, then
        # AutoMinorVersionUpgrade is not in the result dict.
        result = {
            'CacheClusterId': 'cachecluster-us-east-1c',
            'CacheNodeType': 'cache.m1.small',
            'CacheSecurityGroupNames': ['group1', 'group2'],
            'Engine': 'memcached',
            'EngineVersion': '1.4.5',
            'NumCacheNodes': 1,
            'PreferredAvailabilityZone': 'us-east-1c',
            'PreferredMaintenanceWindow': 'fri:08:00-fri:09:00',
        }
        self.assert_params_for_cmd(cmdline, result)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime

import pytest

from awscli.customizations.s3.filegenerator import FileStat
from awscli.customizations.s3.syncstrategy.nooverwrite import NoOverwriteSync


@pytest.fixture
def sync_strategy():
    return NoOverwriteSync('file_at_src_and_dest')


def test_file_exists_at_destination_with_same_key(sync_strategy):
    time_src = datetime.datetime.now()

    src_file = FileStat(
        src='',
        dest='',
        compare_key='test.py',
        size=10,
        last_update=time_src,
        src_type='local',
        dest_type='s3',
        operation_name='upload',
    )
    dest_file = FileStat(
        src='',
        dest='',
        compare_key='test.py',
        size=100,
        last_update=time_src,
        src_type='local',
        dest_type='s3',
        operation_name='',
    )
    should_sync = sync_strategy.determine_should_sync(src_file, dest_file)
    assert not should_sync



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest
from botocore.exceptions import DataNotFoundError
from botocore.model import OperationModel

from awscli.customizations import paginate
from awscli.customizations.exceptions import ParamValidationError
from awscli.customizations.paginate import PageArgument
from awscli.help import OperationDocumentEventHandler, OperationHelpCommand
from awscli.testutils import mock, unittest


@pytest.fixture
def max_items_page_arg():
    return PageArgument('max-items', 'documentation', int, 'MaxItems')


class TestPaginateBase(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock()
        self.paginator_model = mock.Mock()
        self.pagination_config = {
            'input_token': 'Foo',
            'limit_key': 'Bar',
        }
        self.paginator_model.get_paginator.return_value = (
            self.pagination_config
        )
        self.session.get_paginator_model.return_value = self.paginator_model

        self.operation_model = mock.Mock()
        self.foo_param = mock.Mock()
        self.foo_param.name = 'Foo'
        self.foo_param.type_name = 'string'
        self.bar_param = mock.Mock()
        self.bar_param.type_name = 'string'
        self.bar_param.name = 'Bar'
        self.params = [self.foo_param, self.bar_param]
        self.operation_model.input_shape.members = {
            "Foo": self.foo_param,
            "Bar": self.bar_param,
        }


class TestArgumentTableModifications(TestPaginateBase):
    def test_customize_arg_table(self):
        argument_table = {
            'foo': mock.Mock(),
            'bar': mock.Mock(),
        }
        paginate.unify_paging_params(
            argument_table,
            self.operation_model,
            'building-argument-table.foo.bar',
            self.session,
        )
        # We should mark the built in input_token as 'hidden'.
        self.assertTrue(argument_table['foo']._UNDOCUMENTED)
        # Also need to hide the limit key.
        self.assertTrue(argument_table['bar']._UNDOCUMENTED)
        # We also need to inject startin-token and max-items.
        self.assertIn('starting-token', argument_table)
        self.assertIn('max-items', argument_table)
        self.assertIn('page-size', argument_table)
        # And these should be PageArguments.
        self.assertIsInstance(
            argument_table['starting-token'], paginate.PageArgument
        )
        self.assertIsInstance(
            argument_table['max-items'], paginate.PageArgument
        )
        self.assertIsInstance(
            argument_table['page-size'], paginate.PageArgument
        )

    def test_operation_with_no_paginate(self):
        # Operations that don't paginate are left alone.
        self.paginator_model.get_paginator.side_effect = ValueError()
        argument_table = {
            'foo': 'FakeArgObject',
            'bar': 'FakeArgObject',
        }
        starting_table = argument_table.copy()
        paginate.unify_paging_params(
            argument_table,
            self.operation_model,
            'building-argument-table.foo.bar',
            self.session,
        )
        self.assertEqual(starting_table, argument_table)

    def test_service_with_no_paginate(self):
        # Operations that don't paginate are left alone.
        self.session.get_paginator_model.side_effect = DataNotFoundError(
            data_path='foo.paginators.json'
        )
        argument_table = {
            'foo': 'FakeArgObject',
            'bar': 'FakeArgObject',
        }
        starting_table = argument_table.copy()
        paginate.unify_paging_params(
            argument_table,
            self.operation_model,
            'building-argument-table.foo.bar',
            self.session,
        )
        self.assertEqual(starting_table, argument_table)


class TestHelpDocumentationModifications(TestPaginateBase):
    def test_injects_pagination_help_text(self):
        with mock.patch(
            'awscli.customizations.paginate.get_paginator_config',
            return_value={'result_key': 'abc'},
        ):
            help_command = OperationHelpCommand(
                mock.Mock(),
                mock.Mock(),
                mock.Mock(),
                'foo',
                OperationDocumentEventHandler,
            )
            help_command.obj = mock.Mock(OperationModel)
            help_command.obj.name = 'foo'
            paginate.add_paging_description(help_command)
            self.assertIn(
                '``foo`` is a paginated operation. Multiple API',
                help_command.doc.getvalue().decode(),
            )
            self.assertIn(
                'following query expressions: ``abc``',
                help_command.doc.getvalue().decode(),
            )

    def test_shows_result_keys_when_array(self):
        with mock.patch(
            'awscli.customizations.paginate.get_paginator_config',
            return_value={'result_key': ['abc', '123']},
        ):
            help_command = OperationHelpCommand(
                mock.Mock(),
                mock.Mock(),
                mock.Mock(),
                'foo',
                OperationDocumentEventHandler,
            )
            help_command.obj = mock.Mock(OperationModel)
            help_command.obj.name = 'foo'
            paginate.add_paging_description(help_command)
            self.assertIn(
                'following query expressions: ``abc``, ``123``',
                help_command.doc.getvalue().decode(),
            )

    def test_does_not_show_result_key_if_not_present(self):
        with mock.patch(
            'awscli.customizations.paginate.get_paginator_config',
            return_value={'limit_key': 'aaa'},
        ):
            help_command = OperationHelpCommand(
                mock.Mock(),
                mock.Mock(),
                mock.Mock(),
                'foo',
                OperationDocumentEventHandler,
            )
            help_command.obj = mock.Mock(OperationModel)
            help_command.obj.name = 'foo'
            paginate.add_paging_description(help_command)
            self.assertIn(
                '``foo`` is a paginated operation. Multiple API',
                help_command.doc.getvalue().decode(),
            )
            self.assertNotIn(
                'following query expressions',
                help_command.doc.getvalue().decode(),
            )

    def test_does_not_inject_when_no_pagination(self):
        with mock.patch(
            'awscli.customizations.paginate.get_paginator_config',
            return_value=None,
        ):
            help_command = OperationHelpCommand(
                mock.Mock(),
                mock.Mock(),
                mock.Mock(),
                'foo',
                OperationDocumentEventHandler,
            )
            help_command.obj = mock.Mock(OperationModel)
            help_command.obj.name = 'foo'
            paginate.add_paging_description(help_command)
            self.assertNotIn(
                '``foo`` is a paginated operation',
                help_command.doc.getvalue().decode(),
            )


class TestStringLimitKey(TestPaginateBase):
    def setUp(self):
        super(TestStringLimitKey, self).setUp()
        self.bar_param.type_name = 'string'

    def test_integer_limit_key(self):
        argument_table = {
            'foo': mock.Mock(),
            'bar': mock.Mock(),
        }
        paginate.unify_paging_params(
            argument_table,
            self.operation_model,
            'building-argument-table.foo.bar',
            self.session,
        )
        # Max items should be the same type as bar, which may not be an int
        self.assertEqual('string', argument_table['max-items'].cli_type_name)


class TestIntegerLimitKey(TestPaginateBase):
    def setUp(self):
        super(TestIntegerLimitKey, self).setUp()
        self.bar_param.type_name = 'integer'

    def test_integer_limit_key(self):
        argument_table = {
            'foo': mock.Mock(),
            'bar': mock.Mock(),
        }
        paginate.unify_paging_params(
            argument_table,
            self.operation_model,
            'building-argument-table.foo.bar',
            self.session,
        )
        # Max items should be the same type as bar, which may not be an int
        self.assertEqual('integer', argument_table['max-items'].cli_type_name)


class TestBadLimitKey(TestPaginateBase):
    def setUp(self):
        super(TestBadLimitKey, self).setUp()
        self.bar_param.type_name = 'bad'

    def test_integer_limit_key(self):
        argument_table = {
            'foo': mock.Mock(),
            'bar': mock.Mock(),
        }
        with self.assertRaises(TypeError):
            paginate.unify_paging_params(
                argument_table,
                self.operation_model,
                'building-argument-table.foo.bar',
                self.session,
            )


class TestShouldEnablePagination(TestPaginateBase):
    def setUp(self):
        super(TestShouldEnablePagination, self).setUp()
        self.parsed_globals = mock.Mock()
        self.parsed_args = mock.Mock()
        self.parsed_args.starting_token = None
        self.parsed_args.page_size = None
        self.parsed_args.max_items = None
        self.call_parameters = {}

    def test_should_not_enable_pagination(self):
        # Here the user has specified a manual pagination argument,
        # so we should turn pagination off.
        # From setUp(), the limit_key is 'Bar'
        input_tokens = ['foo', 'bar']
        self.parsed_globals.paginate = True
        # Corresponds to --bar 10
        self.parsed_args.foo = None
        self.parsed_args.bar = 10
        paginate.check_should_enable_pagination(
            input_tokens, {}, {}, self.parsed_args, self.parsed_globals
        )
        # We should have turned paginate off because the
        # user specified --bar 10
        self.assertFalse(self.parsed_globals.paginate)

    def test_should_not_enable_pagination_call_parameters(self):
        # Here the user has specified a manual pagination argument,
        # via CLI Input JSON so we should turn pagination off.
        # From setUp(), the limit_key is 'Bar'
        input_tokens = ['Foo', 'Bar']
        self.parsed_globals.paginate = True
        # Corresponds to --bar 10
        self.call_parameters['Foo'] = None
        self.call_parameters['Bar'] = 10
        paginate.check_should_enable_pagination_call_parameters(
            input_tokens, self.call_parameters, {}, self.parsed_globals
        )
        # We should have turned paginate off because the
        # user specified {Bar: 10} in the input JSON
        self.assertFalse(self.parsed_globals.paginate)

    def test_should_enable_pagination_with_no_args(self):
        input_tokens = ['foo', 'bar']
        self.parsed_globals.paginate = True
        # Corresponds to not specifying --foo nor --bar
        self.parsed_args.foo = None
        self.parsed_args.bar = None
        paginate.check_should_enable_pagination(
            input_tokens, {}, {}, self.parsed_args, self.parsed_globals
        )
        # We should have turned paginate off because the
        # user specified --bar 10
        self.assertTrue(self.parsed_globals.paginate)

    def test_should_enable_pagination_with_no_args_call_parameters(self):
        input_tokens = ['Foo', 'Bar']
        self.parsed_globals.paginate = True
        # Corresponds to not specifying --foo nor --bar
        self.call_parameters['Fie'] = 1
        self.call_parameters['Baz'] = 2
        paginate.check_should_enable_pagination_call_parameters(
            input_tokens, self.call_parameters, {}, self.parsed_globals
        )
        # We should have not have turned paginate off because the
        # user specified did not specify any pagination args
        self.assertTrue(self.parsed_globals.paginate)

    def test_default_to_pagination_on_when_ambiguous(self):
        input_tokens = ['foo', 'max-items']
        self.parsed_globals.paginate = True
        # Here the user specifies --max-items 10 This is ambiguous because the
        # input_token also contains 'max-items'.  Should we assume they want
        # pagination turned off or should we assume that this is the normalized
        # --max-items?
        # Will we default to assuming they meant the normalized
        # --max-items.
        self.parsed_args.foo = None
        self.parsed_args.max_items = 10
        paginate.check_should_enable_pagination(
            input_tokens, {}, {}, self.parsed_args, self.parsed_globals
        )
        self.assertTrue(
            self.parsed_globals.paginate, "Pagination was not enabled."
        )

    def test_fall_back_to_original_max_items_when_pagination_turned_off(self):
        input_tokens = ['max-items']
        # User specifies --no-paginate.
        self.parsed_globals.paginate = False
        # But also specifies --max-items 10, which is normally a pagination arg
        # we replace.  However, because they've explicitly turned off
        # pagination, we should put back the original arg.
        self.parsed_args.max_items = 10
        shadowed_args = {'max-items': mock.sentinel.ORIGINAL_ARG}
        arg_table = {'max-items': mock.sentinel.PAGINATION_ARG}

        paginate.check_should_enable_pagination(
            input_tokens,
            shadowed_args,
            arg_table,
            self.parsed_args,
            self.parsed_globals,
        )

    def test_shadowed_args_are_replaced_when_pagination_turned_off(self):
        input_tokens = ['foo', 'bar']
        self.parsed_globals.paginate = True
        # Corresponds to --bar 10
        self.parsed_args.foo = None
        self.parsed_args.bar = 10
        shadowed_args = {'foo': mock.sentinel.ORIGINAL_ARG}
        arg_table = {'foo': mock.sentinel.PAGINATION_ARG}
        paginate.check_should_enable_pagination(
            input_tokens,
            shadowed_args,
            arg_table,
            self.parsed_args,
            self.parsed_globals,
        )
        # We should have turned paginate off because the
        # user specified --bar 10
        self.assertFalse(self.parsed_globals.paginate)
        self.assertEqual(arg_table['foo'], mock.sentinel.ORIGINAL_ARG)

    def test_shadowed_args_are_replaced_when_pagination_set_off(self):
        input_tokens = ['foo', 'bar']
        self.parsed_globals.paginate = False
        # Corresponds to --bar 10
        self.parsed_args.foo = None
        self.parsed_args.bar = 10
        shadowed_args = {'foo': mock.sentinel.ORIGINAL_ARG}
        arg_table = {'foo': mock.sentinel.PAGINATION_ARG}
        paginate.check_should_enable_pagination(
            input_tokens,
            shadowed_args,
            arg_table,
            self.parsed_args,
            self.parsed_globals,
        )
        # We should have turned paginate off because the
        # user specified --bar 10
        self.assertFalse(self.parsed_globals.paginate)
        self.assertEqual(arg_table['foo'], mock.sentinel.ORIGINAL_ARG)


class TestEnsurePagingParamsNotSet(TestPaginateBase):
    def setUp(self):
        super(TestEnsurePagingParamsNotSet, self).setUp()
        self.parsed_args = mock.Mock()

        self.parsed_args.starting_token = None
        self.parsed_args.page_size = None
        self.parsed_args.max_items = None

    def test_pagination_params_raise_error_with_no_paginate(self):
        self.parsed_args.max_items = 100

        with self.assertRaises(ParamValidationError):
            paginate.ensure_paging_params_not_set(self.parsed_args, {})

    def test_can_handle_missing_page_size(self):
        # Not all pagination operations have a page_size.
        del self.parsed_args.page_size
        self.assertIsNone(
            paginate.ensure_paging_params_not_set(self.parsed_args, {})
        )


class TestNonPositiveMaxItems:
    def test_positive_integer_does_not_raise_warning(
        self, max_items_page_arg, capsys
    ):
        max_items_page_arg.add_to_params({}, 1)
        captured = capsys.readouterr()
        assert captured.err == ""

    def test_zero_raises_warning(self, max_items_page_arg, capsys):
        max_items_page_arg.add_to_params({}, 0)
        captured = capsys.readouterr()
        assert "Non-positive values for --max-items" in captured.err

    def test_negative_integer_raises_warning(self, max_items_page_arg, capsys):
        max_items_page_arg.add_to_params({}, -1)
        captured = capsys.readouterr()
        assert "Non-positive values for --max-items" in captured.err



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.datapipeline import ParameterValuesInlineArgument
from awscli.testutils import unittest


class TestParameterValuesInlineArgument(unittest.TestCase):
    def test_over_2_values_with_same_key(self):
        parameters = {}
        argument = ParameterValuesInlineArgument('parameter-values')
        argument.add_to_params(
            parameters,
            [
                'param1=value1',
                'param1=value2',
                'param1=value3',
            ],
        )
        self.assertEqual(
            parameters['parameterValues'],
            [
                {"id": "param1", "stringValue": "value1"},
                {"id": "param1", "stringValue": "value2"},
                {"id": "param1", "stringValue": "value3"},
            ],
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os

from awscli.testutils import BaseAWSCommandParamsTest, FileCreator


class TestAddModel(BaseAWSCommandParamsTest):
    prefix = 'configure add-model'

    def setUp(self):
        super(TestAddModel, self).setUp()
        self.files = FileCreator()
        self.customer_data_root = self.files.rootdir
        self.data_loader = self.driver.session.get_component('data_loader')
        self.data_loader.CUSTOMER_DATA_PATH = self.customer_data_root
        self.service_definition = {
            "version": "2.0",
            "metadata": {
                "apiVersion": '2015-12-02',
                "endpointPrefix": 'myservice',
            },
            "operations": {},
            "shapes": {},
        }
        self.service_unicode_definition = {
            "version": "2.0",
            "metadata": {
                "apiVersion": '2015-12-02',
                "endpointPrefix": 'myservice',
                "keyWithUnicode": '\u2713',
            },
            "operations": {},
            "shapes": {},
        }

    def tearDown(self):
        super(TestAddModel, self).tearDown()
        self.files.remove_all()

    def test_add_model(self):
        cmdline = self.prefix + ' --service-model %s' % json.dumps(
            self.service_definition, separators=(',', ':')
        )
        self.run_cmd(cmdline)

        # Ensure that the model exists in the correct location.
        self.assertTrue(
            os.path.exists(
                os.path.join(
                    self.customer_data_root,
                    'myservice',
                    '2015-12-02',
                    'service-2.json',
                )
            )
        )

    def test_add_model_with_unicode(self):
        cmdline = self.prefix + ' --service-model %s' % json.dumps(
            self.service_unicode_definition, separators=(',', ':')
        )
        self.run_cmd(cmdline)

        # Ensure that the model exists in the correct location.
        self.assertTrue(
            os.path.exists(
                os.path.join(
                    self.customer_data_root,
                    'myservice',
                    '2015-12-02',
                    'service-2.json',
                )
            )
        )

    def test_add_model_with_service_name(self):
        cmdline = self.prefix + ' --service-model %s' % json.dumps(
            self.service_definition, separators=(',', ':')
        )
        cmdline += ' --service-name override-name'
        self.run_cmd(cmdline)

        # Ensure that the model exists in the correct location.
        self.assertTrue(
            os.path.exists(
                os.path.join(
                    self.customer_data_root,
                    'override-name',
                    '2015-12-02',
                    'service-2.json',
                )
            )
        )



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""PEP 517 interface for building the AWS CLI

The interface for the public functions defined in this module are dictated by
PEP 517: https://www.python.org/dev/peps/pep-0517/#build-backend-interface.
Typically projects just rely directly on public build libraries (e.g.,
flit_core) in their pyproject.toml. However, the AWS CLI requires an
auto-complete index to function properly, and instead of directly committing
this index (which is MBs in size and changes with each API update) to the
repository, it is built as part of building the wheel. For the most part,
this in-tree backend just proxies to flit logic. The only exception
is that it builds the auto-complete index and injects it into the wheel
built by flit prior to returning.
"""

import base64
import contextlib
import glob
import hashlib
import os
import re
import shutil
import sys
import tarfile
import zipfile
from pathlib import Path

import flit_core.buildapi

ROOT_DIR = Path(__file__).parents[1]
sys.path.insert(0, str(ROOT_DIR))
BIN_DIR = ROOT_DIR / "bin"
BUILD_DIR = ROOT_DIR / "build"
SCRIPTS = os.listdir(BIN_DIR)
AC_INDEX_REL_PATH = os.path.join("awscli", "data", "ac.index")


def build_sdist(sdist_directory, config_settings=None):
    sdist_filename = flit_core.buildapi.build_sdist(
        sdist_directory, config_settings
    )
    _update_sdist(os.path.join(sdist_directory, sdist_filename))
    return sdist_filename


def build_wheel(
    wheel_directory,
    config_settings=None,
    metadata_directory=None,
):
    whl_filename = flit_core.buildapi.build_wheel(
        wheel_directory, config_settings, metadata_directory
    )
    _inject_wheel_extras(os.path.join(wheel_directory, whl_filename))
    return whl_filename


def build_editable(
    wheel_directory,
    config_settings=None,
    metadata_directory=None,
):
    whl_filename = flit_core.buildapi.build_editable(
        wheel_directory, config_settings, metadata_directory
    )
    _inject_editable_wheel_extras(os.path.join(wheel_directory, whl_filename))
    _build_ac_index(ROOT_DIR / "awscli" / "data", rebuild=False)
    return whl_filename


def prepare_metadata_for_build_wheel(metadata_directory, config_settings=None):
    return flit_core.buildapi.prepare_metadata_for_build_wheel(
        metadata_directory, config_settings
    )


def prepare_metadata_for_build_editable(
    metadata_directory, config_settings=None
):
    return flit_core.buildapi.prepare_metadata_for_build_editable(
        metadata_directory, config_settings
    )


def get_requires_for_build_sdist(config_settings=None):
    return flit_core.buildapi.get_requires_for_build_sdist(config_settings)


def get_requires_for_build_wheel(config_settings=None):
    requires = flit_core.buildapi.get_requires_for_build_wheel(config_settings)
    # Generation of the auto-complete index requires importing from the
    # awscli package and iterating over the commands from the clidriver. In
    # order to be able to do this, it requires all of the CLI's runtime
    # dependencies to be present to avoid import errors.
    dependency_block_re = re.compile(
        r"dependencies = \[([\s\S]+?)\]\s", re.MULTILINE
    )
    extract_dependencies_re = re.compile(r'"(.+)"')
    with open(ROOT_DIR / "pyproject.toml") as f:
        data = f.read()
    raw_dependencies = dependency_block_re.findall(data)[0]
    dependencies = extract_dependencies_re.findall(raw_dependencies)
    return dependencies + requires


get_requires_for_build_editable = get_requires_for_build_wheel


def _update_sdist(sdist_path):
    with _extracted_sdist_dir(sdist_path) as extracted_sdist_dir:
        # Remove the ac.index if it exists. flit backend includes all data
        # files so we need to remove it if it happens to have been built.
        _remove_file_if_exists(
            os.path.join(extracted_sdist_dir, AC_INDEX_REL_PATH)
        )
        _inject_extra_sdist_files(extracted_sdist_dir)


def _inject_extra_sdist_files(tar_root):
    # Flit backend has no way of including extra data files that are not a part
    # of the base package. This adds additional data files that are needed into
    # the sdist.
    for pattern in read_sdist_extras():
        pattern = pattern.replace("/", os.sep)
        pattern = os.path.join(ROOT_DIR, pattern)
        for filename in glob.glob(pattern, recursive=True):
            filename = os.path.relpath(filename, ROOT_DIR)
            path_to_add = os.path.join(ROOT_DIR, filename)
            if not _should_copy(path_to_add):
                continue
            target_path = os.path.join(tar_root, filename)
            _create_dir_if_not_exists(os.path.dirname(target_path))
            shutil.copy2(path_to_add, target_path)


def _should_copy(path):
    if "__pycache__" in path or path.endswith(".pyc"):
        return False
    if os.path.isdir(path):
        return False
    return True


def read_sdist_extras():
    with open(ROOT_DIR / "pyproject.toml") as f:
        data = f.read()
    # This regex searches for the list content of sdist_extra_files
    # in the tool.awscli table within pyproject.toml.
    # We do this to avoid a dependency on tomli until it settles into
    # the stdlib.
    # "# end of cli tool section comment" is to ensure we don't parse
    # something with a similar name elsewhere.
    sdist_extra_block_re = re.compile(
        (
            r"[tool.awscli.sdist].*?"
            r"include = \[([\s\S]+?)\].*?"
            r"# end of cli sdist tool section"
        ),
        re.MULTILINE | re.DOTALL,
    )
    extract_dependencies_re = re.compile(r'"(.+)"')

    raw_extras = sdist_extra_block_re.findall(data)[0]
    extras = extract_dependencies_re.findall(raw_extras)
    return extras


def _rewrite_shebang(path):
    with open(path) as f:
        lines = f.read().split("\n")
    # Rewrite shebang lines to be #!python to conform with PEP 427.
    if lines[0] == "#!/usr/bin/env python":
        lines[0] = "#!python"
    with open(path, "w") as f:
        f.write("\n".join(lines))


def _inject_wheel_extras(whl_path):
    with _extracted_wheel_dir(whl_path) as extracted_wheel_dir:
        _build_and_inject_ac_index(BUILD_DIR, extracted_wheel_dir)
        _inject_scripts(extracted_wheel_dir)


def _inject_editable_wheel_extras(whl_path):
    with _extracted_wheel_dir(whl_path) as extracted_wheel_dir:
        _inject_scripts(extracted_wheel_dir)


@contextlib.contextmanager
def _extracted_wheel_dir(whl_path):
    _create_dir_if_not_exists(BUILD_DIR)
    unpack_wheel_dir = os.path.join(BUILD_DIR, "unpacked_wheel")
    _remove_dir_if_exists(unpack_wheel_dir)
    with _open_wheel(whl_path, unpack_wheel_dir) as extracted_wheel_dir:
        yield extracted_wheel_dir


@contextlib.contextmanager
def _extracted_sdist_dir(sdist_path):
    _create_dir_if_not_exists(BUILD_DIR)
    unpack_sdist_dir = os.path.join(BUILD_DIR, "unpacked_sdist")
    _remove_dir_if_exists(unpack_sdist_dir)
    with _open_sdist(sdist_path, unpack_sdist_dir) as extracted_sdist_dir:
        yield extracted_sdist_dir


def _build_and_inject_ac_index(build_dir, extracted_wheel_dir):
    ac_index_build_name = _build_ac_index(build_dir)
    extracted_ac_index = os.path.join(extracted_wheel_dir, AC_INDEX_REL_PATH)
    _remove_file_if_exists(extracted_ac_index)
    print("Adding auto-complete index into wheel")
    os.rename(
        ac_index_build_name,
        extracted_ac_index,
    )


def _build_ac_index(build_dir, rebuild=True):
    from awscli.autocomplete.generator import generate_index

    ac_index_build_name = os.path.join(build_dir, "ac.index")
    if rebuild:
        _remove_file_if_exists(ac_index_build_name)
    elif os.path.exists(ac_index_build_name):
        return ac_index_build_name

    print("Generating auto-complete index")
    generate_index(ac_index_build_name)
    return ac_index_build_name


def _inject_scripts(extracted_wheel_dir):
    # There are two directories generated by flit:
    # 1) awscli/
    # 2) awscli-version.dist-info/
    # The easiest way to generate a data directory is to
    # find the dist-info directory and replace dist-info with
    # data.
    # There shold be exactly one, if not then something has gone horribly wrong
    # and crashing here is appropriate.
    dist_info_dir = [
        d for d in os.listdir(extracted_wheel_dir) if "dist-info" in d
    ][0]
    data_dir = os.path.join(
        extracted_wheel_dir,
        dist_info_dir.replace("dist-info", "data"),
    )
    _create_dir_if_not_exists(data_dir)
    scripts_dir = os.path.join(data_dir, "scripts")
    _create_dir_if_not_exists(scripts_dir)
    for script in SCRIPTS:
        target = os.path.join(scripts_dir, script)
        print(f"Injecting script {script} -> {target}")
        shutil.copy2(
            os.path.join(BIN_DIR, script),
            target,
        )
        _rewrite_shebang(target)
    _create_record_file(extracted_wheel_dir, dist_info_dir)


def _create_record_file(base_dir, dist_info_dir):
    entries = []
    for base, dirs, files in os.walk(base_dir):
        dirs.sort()
        for filename in sorted(files):
            # Record file has its own special entry at the end
            if filename == "RECORD":
                continue
            target = os.path.join(base, filename)
            # Path entries in record file are unix style
            path = os.path.relpath(target, base_dir).replace(os.sep, "/")
            digest = _get_b64_sha256(target)
            size = os.path.getsize(target)
            record_entry = f"{path},sha256={digest},{size}"
            entries.append(record_entry)
    entries.append(f"{dist_info_dir}/RECORD,,")
    record_path = os.path.join(base_dir, dist_info_dir, "RECORD")
    with open(record_path, "w") as f:
        f.write("\n".join(entries) + "\n")


def _get_b64_sha256(path):
    with open(path, "rb") as f:
        file_bytes = f.read()
        digest_bytes = hashlib.sha256(file_bytes).digest()
    encoded = base64.urlsafe_b64encode(digest_bytes).rstrip(b"=")
    return encoded.decode("utf-8")


@contextlib.contextmanager
def _open_wheel(whl_path, extract_dir="."):
    _extract_zip(whl_path, extract_dir)
    yield extract_dir
    _compress_whl(whl_path, extract_dir)


@contextlib.contextmanager
def _open_sdist(sdist_path, extract_dir="."):
    _extract_sdist(sdist_path, extract_dir)
    basename = os.listdir(extract_dir)[0]
    tar_root = os.path.join(extract_dir, basename)
    yield tar_root
    _compress_sdist(sdist_path, tar_root)


def _extract_sdist(sdist_path, target_dir):
    sdist = tarfile.open(sdist_path)
    sdist.extractall(target_dir)
    sdist.close()


def _compress_sdist(sdist_path, target_dir):
    with tarfile.open(sdist_path, "w:gz") as tar:
        tar.add(target_dir, arcname=os.path.basename(target_dir))


def _extract_zip(zipfile_name, target_dir):
    with zipfile.ZipFile(zipfile_name, "r") as zf:
        for zf_info in zf.infolist():
            # Works around extractall not preserving file permissions:
            # https://bugs.python.org/issue15795
            extracted_path = zf.extract(zf_info, target_dir)
            os.chmod(extracted_path, zf_info.external_attr >> 16)


def _compress_whl(whl_file_path, target_dir):
    _remove_file_if_exists(whl_file_path)
    shutil.make_archive(whl_file_path, "zip", target_dir)
    os.rename(whl_file_path + ".zip", whl_file_path)


def _remove_file_if_exists(filename):
    if os.path.exists(filename):
        os.remove(filename)


def _remove_dir_if_exists(dirname):
    if os.path.isdir(dirname):
        shutil.rmtree(dirname)


def _create_dir_if_not_exists(dirname):
    if not os.path.isdir(dirname):
        os.makedirs(dirname)



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import platform

from awscli.customizations.s3.filegenerator import FileStat
from awscli.customizations.s3.filters import Filter, create_filter
from awscli.testutils import unittest


def platform_path(filepath):
    # Convert posix platforms to windows platforms.
    if platform.system().lower() == 'windows':
        filepath = filepath.replace('/', os.sep)
        filepath = 'C:' + filepath
    return filepath


class FiltersTest(unittest.TestCase):
    def setUp(self):
        self.local_files = [
            self.file_stat('test.txt'),
            self.file_stat('test.jpg'),
            self.file_stat(os.path.join('directory', 'test.jpg')),
        ]
        self.s3_files = [
            self.file_stat('bucket/test.txt', src_type='s3'),
            self.file_stat('bucket/test.jpg', src_type='s3'),
            self.file_stat('bucket/key/test.jpg', src_type='s3'),
        ]

    def file_stat(self, filename, src_type='local'):
        if src_type == 'local':
            filename = os.path.abspath(filename)
            dest_type = 's3'
        else:
            dest_type = 'local'
        return FileStat(
            src=filename,
            dest='',
            compare_key='',
            size=10,
            last_update=0,
            src_type=src_type,
            dest_type=dest_type,
            operation_name='',
        )

    def create_filter(
        self, filters=None, root=None, dst_root=None, parameters=None
    ):
        if root is None:
            root = os.getcwd()
        if filters is None:
            filters = {}
        if dst_root is None:
            dst_root = 'bucket'
        if parameters is not None:
            return create_filter(parameters)
        return Filter(filters, root, dst_root)

    def test_no_filter(self):
        exc_inc_filter = self.create_filter()
        matched_files = list(exc_inc_filter.call(self.local_files))
        self.assertEqual(matched_files, self.local_files)

        matched_files2 = list(exc_inc_filter.call(self.s3_files))
        self.assertEqual(matched_files2, self.s3_files)

    def test_include(self):
        patterns = [['include', '*.txt']]
        include_filter = self.create_filter([['include', '*.txt']])
        matched_files = list(include_filter.call(self.local_files))
        self.assertEqual(matched_files, self.local_files)

        matched_files2 = list(include_filter.call(self.s3_files))
        self.assertEqual(matched_files2, self.s3_files)

    def test_exclude(self):
        exclude_filter = self.create_filter([['exclude', '*']])
        matched_files = list(exclude_filter.call(self.local_files))
        self.assertEqual(matched_files, [])

        matched_files = list(exclude_filter.call(self.s3_files))
        self.assertEqual(matched_files, [])

    def test_exclude_with_dst_root(self):
        exclude_filter = self.create_filter(
            [['exclude', '*.txt']], dst_root='bucket'
        )
        matched_files = list(exclude_filter.call(self.local_files))
        b = os.path.basename
        self.assertNotIn('test.txt', [b(f.src) for f in matched_files])
        # Same filter should match the dst files.
        matched_files = list(exclude_filter.call(self.s3_files))
        self.assertNotIn('test.txt', [b(f.src) for f in matched_files])

    def test_exclude_include(self):
        patterns = [['exclude', '*'], ['include', '*.txt']]
        exclude_include_filter = self.create_filter(patterns)
        matched_files = list(exclude_include_filter.call(self.local_files))
        self.assertEqual(matched_files, [self.local_files[0]])

        matched_files = list(exclude_include_filter.call(self.s3_files))
        self.assertEqual(matched_files, [self.s3_files[0]])

    def test_include_exclude(self):
        patterns = [['include', '*.txt'], ['exclude', '*']]
        exclude_all_filter = self.create_filter(patterns)
        matched_files = list(exclude_all_filter.call(self.local_files))
        self.assertEqual(matched_files, [])

        matched_files = list(exclude_all_filter.call(self.s3_files))
        self.assertEqual(matched_files, [])

    def test_prefix_filtering_consistent(self):
        # The same filter should work for both local and remote files.
        # So if I have a directory with 2 files:
        local_files = [
            self.file_stat('test1.txt'),
            self.file_stat('nottest1.txt'),
        ]
        # And the same 2 files remote (note that the way FileStat objects
        # are constructed, we'll have the bucket name but no leading '/'
        # character):
        remote_files = [
            self.file_stat('bucket/test1.txt', src_type='s3'),
            self.file_stat('bucket/nottest1.txt', src_type='s3'),
        ]
        # If I apply the filter to the local to the local files.
        exclude_filter = self.create_filter([['exclude', 't*']])
        filtered_files = list(exclude_filter.call(local_files))
        self.assertEqual(len(filtered_files), 1)
        self.assertEqual(
            os.path.basename(filtered_files[0].src), 'nottest1.txt'
        )

        # I should get the same result if I apply the same filter to s3
        # objects.
        exclude_filter = self.create_filter([['exclude', 't*']], root='bucket')
        same_filtered_files = list(exclude_filter.call(remote_files))
        self.assertEqual(len(same_filtered_files), 1)
        self.assertEqual(
            os.path.basename(same_filtered_files[0].src), 'nottest1.txt'
        )

    def test_bucket_exclude_with_prefix(self):
        s3_files = [
            self.file_stat('bucket/dir1/key1.txt', src_type='s3'),
            self.file_stat('bucket/dir1/key2.txt', src_type='s3'),
            self.file_stat('bucket/dir1/notkey3.txt', src_type='s3'),
        ]
        filtered_files = list(
            self.create_filter([['exclude', 'dir1/*']], root='bucket').call(
                s3_files
            )
        )
        self.assertEqual(filtered_files, [])

        key_files = list(
            self.create_filter([['exclude', 'dir1/key*']], root='bucket').call(
                s3_files
            )
        )
        self.assertEqual(len(key_files), 1)
        self.assertEqual(key_files[0].src, 'bucket/dir1/notkey3.txt')

    def test_root_dir(self):
        p = platform_path
        local_files = [self.file_stat(p('/foo/bar/baz.txt'), src_type='local')]
        local_filter = self.create_filter(
            [['exclude', 'baz.txt']], root=p('/foo/bar/')
        )
        filtered = list(local_filter.call(local_files))
        self.assertEqual(filtered, [])

        # However, if we're at the root of /foo', then this filter won't match.
        local_filter = self.create_filter(
            [['exclude', 'baz.txt']], root=p('/foo/')
        )
        filtered = list(local_filter.call(local_files))
        self.assertEqual(len(filtered), 1)
        self.assertEqual(filtered[0].src, p('/foo/bar/baz.txt'))

    def test_create_root_s3_with_prefix(self):
        parameters = {
            'filters': [['--exclude', 'test.txt']],
            'dir_op': True,
            'src': 's3://bucket/prefix/',
            'dest': 'prefix',
        }
        s3_filter = self.create_filter(parameters=parameters)
        s3_files = [
            self.file_stat('bucket/prefix/test.txt', src_type='s3'),
            self.file_stat('bucket/prefix/test2.txt', src_type='s3'),
        ]
        filtered = list(s3_filter.call(s3_files))
        self.assertEqual(len(filtered), 1)
        self.assertEqual(filtered[0].src, 'bucket/prefix/test2.txt')

    def test_create_root_s3_no_dir_op(self):
        parameters = {
            'filters': [['--exclude', 'test.txt']],
            'dir_op': False,
            'src': 's3://bucket/test.txt',
            'dest': 'temp',
        }
        s3_filter = self.create_filter(parameters=parameters)
        s3_files = [
            self.file_stat('bucket/test.txt', src_type='s3'),
        ]
        filtered = list(s3_filter.call(s3_files))
        self.assertEqual(len(filtered), 0)

    def test_create_filter_s3_to_s3(self):
        source = 'bucket/'
        destination = 'bucket-2/'
        pattern = '*'
        parameters = {
            'filters': [['--exclude', pattern], ['--include', '*.jpg']],
            'dir_op': True,
            'src': 's3://' + source,
            'dest': 's3://' + destination,
        }
        s3_filter = self.create_filter(parameters=parameters)

        source_pattern = s3_filter.patterns[0][1]
        destination_pattern = s3_filter.dst_patterns[0][1]
        self.assertEqual(source_pattern, source + pattern)
        self.assertEqual(destination_pattern, destination + pattern)

        filtered = list(s3_filter.call(self.s3_files))
        self.assertEqual(len(filtered), 2)
        for filtered_file in filtered:
            self.assertFalse('.txt' in filtered_file.src)


if __name__ == "__main__":
    unittest.main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import shutil
import tempfile

from awscli.compat import ensure_text_type
from awscli.customizations.wizard import ui
from awscli.testutils import cd, mock, unittest


class TestFileCompleter(unittest.TestCase):
    def setUp(self):
        self.temporary_directory = tempfile.mkdtemp()
        self.completer = ui.FileCompleter()

    def tearDown(self):
        shutil.rmtree(self.temporary_directory)

    def touch_file(self, filename):
        with open(os.path.join(self.temporary_directory, filename), 'w'):
            pass

    def get_completions_given_user_input(self, user_input):
        user_input = mock.Mock(text=ensure_text_type(user_input))

        with cd(self.temporary_directory):
            completions = list(
                self.completer.get_completions(user_input, None)
            )
            return completions

    def test_can_list_children_of_dir_with_prefix(self):
        self.touch_file('foo.txt')
        self.touch_file('bar.txt')
        self.touch_file('baz.txt')

        completions = self.get_completions_given_user_input('./b')
        self.assertEqual(
            [c.text for c in completions],
            [os.path.join('.', 'bar.txt'), os.path.join('.', 'baz.txt')],
        )

    def test_full_path_included_if_full_path_used_as_text(self):
        self.touch_file('foo.txt')

        completions = self.get_completions_given_user_input(
            self.temporary_directory + '/'
        )
        self.assertEqual(
            [c.text for c in completions],
            [os.path.join(self.temporary_directory, 'foo.txt')],
        )

    def test_file_not_exists_returns_empy_list(self):
        completions = self.get_completions_given_user_input(
            self.temporary_directory + '/asdf/does-not-exist'
        )
        self.assertEqual(completions, [])
